"""Command-line interface application module

This module provides multiple classes to create compatible CLI tools:
- CommandLineSource
- CommandLineSink
- CommandLinePipe
- CommandLineDeviceSource
- CommandLineDeviceSink
- CommandLineDevicePipe

Standard command-line application classes
-----------------------------------------

Basically, a `CommandLineSource` application can see its standard output tied
to a `CommandLineSink` or  `CommandLinePipe` application, while a  `CommandLineSink`
application can see its standard input fed by a `CommandLineSink` or a
`CommandLinePipe` application. Chosing the right class for an application is
critical because our CLI framework performs some checks and does not allow
for instance a `CommandLineSource` application to see its standard input fed
with another command-line tool.

Generally, applications providing an interactive shell are based on the
`CommandLineSource` application class, as the user is expected to type
commands and to interact with the application's interactive shell.

The following schemes are then possible:

- [CommandLineSource] | [CommandLineSink]
- [CommandLineSource] | [CommandLinePipe] | [CommandLineSink]


WHAD-based command-line application classes
-------------------------------------------

WHAD offers the possibility to chain different tools to create flexible packet
processing flows. This is automatically handled by our CLI device-based application
classes.

An application using the `CommandLineDeviceSource` class will therefore be in
charge of configuring a WHAD adapter specified with the `--interface` option,
perform some actions and then allow a chained tool to take control of this
adapter to perform other actions.

An application using the `CommandLineDeviceSink` class is supposed to get
messages from the previous tool and process them before sending them to a
specific device handled by this application itself. Everything is automated
by our CLI framework and the application's `input_interface` property does
expose the input interface the application is connected to.

Last but not least, an application using the `CommandLineDevicePipe` class will
process every incoming message and forward them (or not) to a chained tool.

"""
import os
import sys
import select
import fcntl
import logging
import traceback

from typing import Generator
from urllib.parse import urlparse, parse_qsl
from signal import signal, SIGPIPE, SIG_DFL
from argparse import ArgumentParser, Namespace

from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.output import create_output
from prompt_toolkit.application.current import get_app_session

from whad.version import get_version
from whad.device import Device, UnixSocket
from whad.exceptions import WhadDeviceAccessDenied, WhadDeviceNotFound, \
    WhadDeviceNotReady, WhadDeviceTimeout, UnsupportedDomain

logger = logging.getLogger(__name__)

signal(SIGPIPE,SIG_DFL)

# Python logging level aliases
LOGLEVEL_ALIASES = {
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'info': logging.INFO,
    'debug': logging.DEBUG
}

class ApplicationError(Exception):
    """Application error
    """
    def __init__(self, reason: str):
        """Initialize an application error.
        """
        super().__init__()
        self.__reason = reason

    def show(self):
        """Show the exception in terminal
        """
        print_formatted_text(HTML(f"<ansired>[!] <b>{self.__reason}</b></ansired>"))

#pylint: disable-next=invalid-name
class command:
    """CommandLineApp command decorator.

    This decorator must be used to register a specific command in the main
    application. You need to provide a short description and a detailed
    description of the command (HTML allowed).
    """

    def __init__(self, cmd_name, short_desc=None, desc=None, category='Generic commands'):
        self.cmd_name = cmd_name
        self.short_desc = short_desc
        self.desc = desc
        self.category = category

    def __call__(self, handler):
        CommandsRegistry.register(
            self.cmd_name,
            handler,
            self.category
        )

class CommandsRegistry:
    """Static registry used to keep track of defined commands and their
    associated documentation.
    """

    COMMANDS = {}
    CMDS_SHORT_DESC = {}
    CMDS_DESC = {}
    CATEGORIES = {}

    @staticmethod
    def register(name: str, handler, category: str):
        """Register a command
        """
        # Add command to our list of known commands
        CommandsRegistry.COMMANDS[name] = handler

        # Add command to the corresponding category
        if category not in CommandsRegistry.CATEGORIES:
            CommandsRegistry.CATEGORIES[category] = []
        CommandsRegistry.CATEGORIES[category].append(name)

        # Extract short desc and description from docstring
        if hasattr(handler, '__doc__'):
            docstr = getattr(handler, '__doc__')
            short_desc = docstr.splitlines()[0].lstrip()
            desc = '\n'.join([l.lstrip() for l in docstr.splitlines()[1:]])
        else:
            short_desc = ''
            desc = ''

        if short_desc is not None:
            CommandsRegistry.CMDS_SHORT_DESC[name] = short_desc
        if desc is not None:
            CommandsRegistry.CMDS_DESC[name] = desc

    @staticmethod
    def get_handler(command_name: str):
        """Get command handler from command name.

        :param command_name: Command name
        :type command_name: str
        :return: function associated with the command or None if none registered
        """
        if command_name in CommandsRegistry.COMMANDS:
            return CommandsRegistry.COMMANDS[command_name]
        return None

    @staticmethod
    def get_short_desc(command_name: str) -> str:
        """Get command short description from command name.

        :param command_name: Command name
        :type command_name: str
        :return: Short description associated with the command or None if none registered
        """
        if command_name in CommandsRegistry.CMDS_SHORT_DESC:
            return CommandsRegistry.CMDS_SHORT_DESC[command_name]
        return None

    @staticmethod
    def get_desc(command_name: str):
        """Get command long description from command name.

        :param command_name: Command name
        :type command_name: str
        :return: Long description associated with the command or None if none registered
        """
        if command_name in CommandsRegistry.CMDS_DESC:
            return CommandsRegistry.CMDS_DESC[command_name]
        return None

    @staticmethod
    def enumerate_categories() -> Generator[str, None, None]:
        """Enumerate over command categories
        """
        yield from CommandsRegistry.CATEGORIES

    @staticmethod
    def enumerate(category=None) -> Generator[tuple, None, None]:
        """Enumerate over all categories or one specific, if provided.

        :param category: Category to iterate
        :type category: str
        """
        if category is None:
            for cmd_name in CommandsRegistry.COMMANDS:
                yield (
                    cmd_name,
                    CommandsRegistry.get_short_desc(cmd_name),
                    CommandsRegistry.get_desc(cmd_name)
                )
        else:
            for cmd_name in CommandsRegistry.COMMANDS:
                if cmd_name in CommandsRegistry.CATEGORIES[category]:
                    yield (
                        cmd_name,
                        CommandsRegistry.get_short_desc(cmd_name),
                        CommandsRegistry.get_desc(cmd_name)
                    )


@command('help')
def show_default_help(_, args):
    """show this help screen

    <ansimagenta><b>help</b> <i>[command]</i></ansimagenta>

    Show contextual help for the provided <i>command</i>
    """
    if len(args) == 0:
        # Enumerate commands by group
        for category in CommandsRegistry.enumerate_categories():
            print_formatted_text(HTML(f"<ansimagenta><b>{category}:</b></ansimagenta>"))
            commands = []
            for cmd, short_desc, _ in CommandsRegistry.enumerate(category):
                commands.append((cmd, short_desc))
            commands.sort()

        # Compute the longest command
        max_cmd_size = max([len(cmd) for cmd,doc in commands])
        cmd_fmt = "<ansicyan>{0:<%d}</ansicyan>\t\t{1}" % max_cmd_size
        for cmd, doc in commands:
            if cmd != 'interactive':
                print_formatted_text(HTML(cmd_fmt.format(cmd, doc)))
        print('')
    else:
        cmd = args[0]
        cmd_desc = CommandsRegistry.get_desc(cmd)
        print_formatted_text(HTML(cmd_desc.strip()))

class CommandLineApp(ArgumentParser):

    """This class provides a wrap-up for WHAD CLI applications, adding
    a common default options:

     --interface/-i: specifies the WHAD interface to use
     --no-color: tells the application not to use colors in terminal

    It also provides a unified way to handle error/warning/info display,
    as well as a convenient way to add specific commands and associated
    handlers.

    Application arguments are expected to follow this pattern:

    $ app [--option [,--option]] [command] [command_arg [, command_arg]]

    Each command is associated to a specific handler that will take care of
    command arguments.
    """
    # WHAD interface related errors
    DEV_READY = 0
    DEV_NOT_FOUND_ERR = -1
    DEV_NOT_READY_ERR = -2
    DEV_ACCESS_ERR = -3

    # Application input types (when piped)
    INPUT_NONE = 0
    INPUT_STANDARD = 1
    INPUT_WHAD = 2

    # Application output types (when piped)
    OUTPUT_NONE = 0
    OUTPUT_STANDARD = 1
    OUTPUT_WHAD = 2

    # Unique application instance
    instance = None

    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 input: int = INPUT_WHAD,
                 output: int = OUTPUT_WHAD, **kwargs):
        """Instantiate a CommandLineApp

        :param str program_name: program (app) name
        :param str usage: usage string
        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        :param int input: specify the input type when application has its stdin piped
        :param int output: specify the output type when application has its stdout piped
        """
        # Save application instance as a class property
        if CommandLineApp.instance is None:
            CommandLineApp.instance = self

        super().__init__(description=description, **kwargs)

        self.__interface = None
        self.__input_iface = None
        self.__input_type = input
        self.__output_type = output
        self.__args = Namespace()
        self.__has_interface = interface
        self.__is_interface_piped = False
        self.__has_commands = commands

        # Add our default option --interface/-i
        if self.__has_interface:
            self.add_argument(
                '--interface', '-i',
                dest='interface',
                help='specifies the WHAD interface to use',
            )

        # Add our default option --no-colorfself
        self.add_argument(
            '--no-color',
            dest='nocolor',
            action='store_true',
            default=False,
            help='disable colors in output'
        )

        # Add our default option --log
        # (-l option not available, we must keep with the long name)
        self.add_argument(
            '--log',
            dest='loglevel',
            choices=['error', 'warn', 'info', 'debug'],
            default=None,
            help='set the logging level for this application'
        )

        # Logfile (-f option not available, we must keep with the long name)
        self.add_argument(
            '--log-file',
            dest='logfile',
            metavar='LOGFILE_PATH',
            default=None,
            help='write the log output into LOGFILE_PATH'
        )

        self.add_argument(
            '--version',
            action='version',
            version=f"%(prog)s {get_version()}"
        )

        # Save application type
        if self.__has_commands:
            self.add_argument(
                'command',
                metavar='COMMAND',
                nargs='?',
                help="command to execute, use 'help' for a list of supported commands"
            )
            self.add_argument(
                'command_args',
                metavar='COMMAND_ARG',
                nargs='*',
                help="command arguments"
            )

    @staticmethod
    def get_instance():
        """Return unique application instance
        """
        return CommandLineApp.instance


    @property
    def interface(self):
        """Return the selected WHAD interface.

        :return WhadDevice: Selected WHAD interface if any, None otherwise.
        """
        return self.__interface


    @interface.setter
    def interface(self, iface):
        """Sets the selected WHAD interface.
        """
        self.__interface = iface

    @property
    def input_interface(self):
        """Input interface
        """
        return self.__input_iface

    @property
    def args(self) -> Namespace:
        """Return the parsed arguments Namespace.
        """
        return self.__args

    def is_piped_interface(self) -> bool:
        """Determine if the input interface is piped from a previous app in
        the command line
        """
        return self.__is_interface_piped

    def pre_run(self):
        """Prepare run for this application

        - parses arguments
        - handling color settings
        - resolve WHAD interface
        - handles piped interfaces
        """
        # First we need to parse the main arguments
        self.__args = self.parse_args()

        # Handle debug options if provided
        if self.__args.loglevel is not None:
            # Convert level to corresponding Python logging level
            desired_level = LOGLEVEL_ALIASES[self.__args.loglevel]

            # Shall we log into a specific file ?
            if self.__args.logfile is not None:
                try:
                    logging.basicConfig(filename=self.__args.logfile,
                                        level=desired_level)
                except IOError:
                    self.error((
                        f"Specified log output file ({self.__args.logfile}) "
                        "cannot be accessed, falling back to stderr."
                    ))
                    logging.basicConfig(level=desired_level)
            else:
                # Else we log into stderr
                logging.basicConfig(level=desired_level)

        # If no color is enabled, change color depth to 1 (black/white)
        if self.__args.nocolor:
            os.environ['PROMPT_TOOLKIT_COLOR_DEPTH']='DEPTH_1_BIT'

        # If interface is provided, instantiate it and make it available
        if self.__has_interface:
            if self.__args.interface is not None:
                try:
                    # Create WHAD interface
                    self.__interface = Device.create(self.__args.interface)
                except WhadDeviceNotFound as dev_404:
                    raise ApplicationError(
                        f"Whad adapter '{self.__args.interface}' not found."
                    ) from dev_404
                except WhadDeviceAccessDenied as dev_403:
                    raise ApplicationError(
                        "Cannot access WHAD device, please check permissions."
                    ) from dev_403
                except WhadDeviceNotReady as dev_500:
                    raise ApplicationError("WHAD device is not ready.") from dev_500

        # If stdout is piped, then tell prompt-toolkit to fallback to another
        # TTY (normally, stderr).
        if self.is_stdout_piped():
            # Check if this application is supposed to be chained to another app
            if self.__output_type != CommandLineApp.OUTPUT_NONE:
                get_app_session()._output = create_output(always_prefer_tty=True)
            else:
                logger.error("This application (%s) cannot be piped with another tool.", self.prog)
                sys.exit(2)

        # If application stdin is piped, we have two possibilities:
        # - the application is supposed to handle "normal" data from stdin
        # - the application is supposed to be chained with another WHAD CLI tool
        if self.is_stdin_piped():
            self.__input_iface = None
            # If stdin is piped to a WHAD tool, we must wait for a specific URL
            # sent in a single line that describes the interface to use.
            if self.__input_type == CommandLineApp.INPUT_WHAD:

                # Set stdin non-blocking
                orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
                fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl | os.O_NONBLOCK)

                # Wait an url on stdin to continue
                logger.debug('Stdin is piped, wait for a valid URL describing our interface.')
                pending_input = ''
                while True:
                    # Check if something is available
                    readers, _, errors = select.select([sys.stdin], [], [sys.stdin], .01)

                    # Check if an error occurred.
                    if len(errors) > 0:
                        # We're done
                        break

                    # Read input from stdin
                    if len(readers) > 0:
                        data_read = sys.stdin.read(4096)
                        if data_read == '' or data_read is None:
                            break

                        pending_input += data_read
                        # Check if we have a full line
                        if '\n' in pending_input:
                            idx = pending_input.index('\n')
                            line, pending_input = pending_input[:idx], pending_input[idx+1:]

                            # parse URL

                            url_info = urlparse(line)
                            if url_info.scheme == 'unix' and url_info.path is not None:
                                self.__interface_path = line
                                self.__is_interface_piped = True

                                # Create a Unix socket device and connect it to the
                                # given Unix socket path
                                self.__input_iface = UnixSocket(url_info.path)
                                self.__input_iface.open()

                                # Copy parameters into our app parameters
                                params = dict(parse_qsl(url_info.query))
                                for param, value in params.items():
                                    if not hasattr(self.args, param):
                                        setattr(self.args, param, value)

                                # We're done
                                break

                            # If stdout is not piped, print line
                            if not self.is_stdout_piped():
                                print(line)
            else:
                # Check if this application is supposed to be chained with
                # another app and to process its standard input
                if self.__input_type == CommandLineApp.INPUT_NONE:
                    logger.error(
                        "This application (%s) cannot be piped with another tool.",
                        self.prog
                    )
                    sys.exit(2)

        # If we have no WHAD adapter specified then that's abnormal.
        elif self.__interface is None and self.__has_interface:
            raise ApplicationError("You must select a WHAD adapter with the --interface option.")


    def post_run(self):
        """Implement post-run tasks.
        """
        if self.__output_type == CommandLineApp.OUTPUT_WHAD:
            # If stdout is piped, forward socket info to next tool
            if isinstance(self.__input_iface, UnixSocket) and self.is_stdout_piped():
                sys.stdout.write(f"{self.__interface_path}\n")
                sys.stdout.flush()
        elif self.__input_type == CommandLineApp.INPUT_WHAD and self.__input_iface is not None:
            # if stdin is piped, close unix socket
            self.__input_iface.close()


    def run(self, pre: bool = True, post: bool = True):
        """Run the main application
        """
        # Launch pre-run tasks if required
        if pre:
            self.pre_run()

        # If we support first positional arg as command, parse the command
        if self.__has_commands:
            if self.__args.command is not None:
                cmd = self.__args.command
                handler = CommandsRegistry.get_handler(cmd)
                if handler is not None:
                    handler(self, self.__args.command_args)
                    return
            elif CommandsRegistry.get_handler('interactive'):
                # If no command is passed to the CLI tool and an interactive
                # command handler has been defined, call it.
                handler = CommandsRegistry.get_handler('interactive')
                if handler is not None:
                    handler(self, self.__args.command_args)
                    return

            # By default, print help if no script is specified
            self.print_help()

        # Launch post-run tasks
        if post:
            self.post_run()


    def is_stdout_piped(self):
        """Checks if stdout is piped to another process

        :return bool: True if stdout is piped, False otherwise
        """
        if not sys.stdout.isatty():
            # If we are on a Linux system, check if we are redirected to
            # /dev/null, in this case we consider we are not *really* piped
            # with a program.
            if sys.platform in ("linux", "linux2"):
                pid = os.getpid()
                stdout_target = os.readlink(f"/proc/{pid}/fd/1")
                if stdout_target != "/dev/null":
                    logger.debug("[cli] stdout is redirected to %s", stdout_target)
                    return True
                else:
                    logger.debug("[cli] stdout redirected to /dev/null, no unix socket needed")
                    return False
            else:
                logger.debug("[cli] stdout is redirected and cannot determine if it leads to /dev/null or not (OS not Linux)")
                return True

        return False


    def is_stdin_piped(self):
        """Checks if stdin is piped by another process

        :return bool: True if stdin is piped, False otherwise
        """
        return not sys.stdin.isatty()

    def readline_from_stdin(self):
        """Reads a line from stdin when application has its stdin piped

        :return str: line read from stdin
        """
        pending_input = ''
        while True:
            data_read = sys.stdin.read(4096)
            if data_read == '' or data_read is None:
                break
        return pending_input

    def warning(self, message, *args):
        """Display a warning message in orange (if color is enabled)
        """
        try:
            print_formatted_text(HTML("<aaa fg=\"#e97f11\">/!\\ <b>{message}</b></aaa>").format(message=message % args))
        except:
            print_formatted_text(HTML(
                "<aaa fg=\"#e97f11\">/!\\ <b>an unknown warning occurred</b></aaa>"
            ))

    def error(self, message, *args):
        """Display an error message in red (if color is enabled)
        """
        try:
            print_formatted_text(
                HTML('<ansired>[!] <b>{message}</b></ansired>').format(message=message % args)
            )
        except Exception as err:
            logger.error('[!] an unknown error occurred: %s',err)

class CommandLineDualOutputApp(CommandLineApp):
    """
    Command-line application that can send data to standard output as well
    as to other chained WHAD command-line tools.

    By default, this class behaves as a `CommandLineApp` application and will
    try to determine by itself if its standard output is piped to another application
    or prints to the standard output.

    This behavior can be overriden by using a specific option, `--force-stdout`, that forces
    the application to send its output to the standard output instead of a chained tool.
    """
    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 input: int = CommandLineApp.INPUT_WHAD,
                 output: int = CommandLineApp.OUTPUT_WHAD, **kwargs):
        # Initialize with parent "constructor".
        super().__init__(description=description, commands=commands, interface=interface,
                         input=input, output=output, **kwargs)

        # Add our default option --stdout
        self.add_argument(
            '--force-stdout',
            dest='stdout_forced',
            action='store_true',
            help='disable tool chaining and force application to print to standard outut'
        )

    def is_stdout_piped(self) -> bool:
        """Override CommandLineApp's default `is_stdout_piped()` method to force the app
        to consider its output not piped, if requested by the user."""
        if self.args.stdout_forced:
            return False
        return super().is_stdout_piped()

class CommandLineSource(CommandLineApp):
    """
    Command-line application that can send data to standard output.
    """
    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that supports no standard
        input processing and is able to send data to standard output for tool
        chaining.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD
                               interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_NONE,
                         CommandLineApp.OUTPUT_STANDARD, **kwargs)


class CommandLineSink(CommandLineApp):
    """
    Command-line application that can read data from standard input.
    """
    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that only supports standard
        input processing and cannot be chained with another tool.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_STANDARD,
                         CommandLineApp.OUTPUT_NONE, **kwargs)

class CommandLinePipe(CommandLineApp):
    """
    Command-line application that reads data from standard input and send data
    to standard output.
    """

    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that supports standard input
        processing and is able to send data to standard output for tool chaining.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_STANDARD,
                         CommandLineApp.OUTPUT_STANDARD, **kwargs)

class CommandLineDeviceSource(CommandLineApp):
    """
    Command-line application that can be chained with a `CommandLineDeviceSink`
    or `CommandLineDevicePipe` application, in a timely manner.
    """

    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that supports no standard
        input processing and is able to send data to standard output for tool
        chaining.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_NONE,
                         CommandLineApp.OUTPUT_WHAD, **kwargs)


class CommandLineDeviceSink(CommandLineApp):
    """
    Command-line application that can receive WHAD messages from a chained source/pipe but cannot be
    chained with another application.
    """

    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that only supports standard
        input processing and cannot be chained with another tool.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_WHAD,
                         CommandLineApp.OUTPUT_NONE, **kwargs)

class CommandLineDevicePipe(CommandLineApp):
    """
    Command-line application that handles/processes WHAD messages between two
    WHAD command-line applications.
    """

    def __init__(self, description: str = None, commands: bool=True, interface: bool=True,
                 **kwargs):
        """Create a command-line application object that supports standard input
        processing and is able to send data to standard output for tool chaining.

        :param str description: program description
        :param bool commands: if enabled, the application will consider first
                              positional argument as a command
        :param bool interface: if enabled, the application will resolve a WHAD interface
        """
        super().__init__(description, commands, interface, CommandLineApp.INPUT_WHAD,
                         CommandLineApp.OUTPUT_WHAD, **kwargs)

def run_app(application: CommandLineApp):
    """Run an application and handle generic exceptions.
    """
    try:
        application.run()
    except ApplicationError as err:
        # If an error occurred, display it.
        err.show()
    except KeyboardInterrupt:
        application.warning("Interrupted by user (CTL-C)")
    except WhadDeviceTimeout:
        application.error("WHAD interface has timed out.")
    except WhadDeviceNotReady:
        application.error("WHAD interface seems unresponsive.")
    except WhadDeviceAccessDenied:
        application.error("Cannot access WHAD interface, check permissions.")
    except UnsupportedDomain as domain_err:
        application.error(f"WHAD interface does not support {domain_err.domain}.")
    except Exception as exc:
        application.error("An unexpected exception occurred:")
        traceback.print_exception(exc)

if __name__ == '__main__':

    @command('test', 'this is a small test command', 'detailed desc')
    def test_handler(_, args):
        """Handle test command
        """
        print(f"this is a test with args: {args}")

    app = CommandLineApp(description='Whad CLI demo')
    app.run()



"""Command-line interface UI helpers.
"""
import sys
import json
from typing import Optional

from hexdump import hexdump
from prompt_toolkit import print_formatted_text, HTML
from scapy.all import Packet, tcpdump , conf


def success(message):
    """Display a success message in green (if color is enabled)
    """
    print_formatted_text(HTML(f"<aaa fg=\"#027923\"><b>{message}</b></aaa>"))

def warning(message):
    """Display a warning message in orange (if color is enabled)
    """
    print_formatted_text(HTML(f"<aaa fg=\"#e97f11\">/!\\ <b>{message}</b></aaa>"))

def error(message):
    """Display an error message in red (if color is enabled)
    """
    print_formatted_text(HTML('<ansired>[!] <b>{message}</b></ansired>').format(message=message))

def info(message):
    """Display an error info message in cyan (if color is enabled)
    """
    print_formatted_text(HTML(f"<ansicyan>[!] <b>{message}</b></ansicyan>"))

pkts = []

def display_packet(pkt, show_metadata=True, format='repr'):
    """
    Display an packet according to the selected format.

    Four main types of formats can be used:
        * repr: scapy packet repr method (default)
        * show: scapy show method, "field" representation
        * hexdump: hexdump representation of the packet content
        * tshark: tshark representation of the packet
        * raw: raw received bytes

    :param  pkt:        Received Signal Strength Indicator
    :type   pkt:        :class:`scapy.packet.packet`
    """
    if isinstance(show_metadata, str):
        show_metadata = show_metadata == 'True'
    if isinstance(pkt, Packet):
        metadata = ""
        if hasattr(pkt, "metadata") and show_metadata:
            metadata = repr(pkt.metadata)
        # Process scapy show method format
        if format == "show":

            if show_metadata:
                print_formatted_text(
                    HTML("<b>{metadata}</b>").format(metadata=metadata)
                )
            pkt.show()

            if hasattr(pkt, "decrypted"):
                print_formatted_text(
                    HTML("<ansicyan>[i] Decrypted payload:</ansicyan>")
                )
                pkt.decrypted.show()

        # Process raw bytes format
        elif format == "raw":

            if show_metadata:
                print_formatted_text(
                    HTML("<b>{metadata}</b>").format(
                        metadata=metadata
                    )
                )

            print(bytes(pkt).hex())

            if hasattr(pkt, "decrypted"):
                print_formatted_text(
                    HTML("<ansicyan>[i] Decrypted payload:</ansicyan> {pkthex}").format(
                        pkthex=bytes(pkt.decrypted).hex()
                    )
                )

        # Process hexdump format
        elif format == "hexdump":
            if show_metadata:
                print_formatted_text(
                    HTML("<b>{metadata}</b>").format(metadata=metadata)
                )
            print_formatted_text(
                HTML("<i>{pkthex}</i>").format(pkthex=hexdump(bytes(pkt), result="return"))
            )
            if hasattr(pkt, "decrypted"):
                print_formatted_text(
                    HTML(
                        "<ansicyan>[i] Decrypted payload:</ansicyan>"
                    )
                )
                print_formatted_text(
                    HTML("<i>{pkthex}</i>").format(
                        pkthex=hexdump(bytes(pkt.decrypted), result="return"))
                )
        elif format == "tshark":
            pkts.append(pkt)
            tcpdump_line = tcpdump(pkts, prog=conf.prog.tshark, dump=True)[:-1].decode('utf-8')
            print(tcpdump_line.split("\n")[-1], end="")
        # Process scapy repr format
        elif format == "repr":
            if show_metadata:
                print_formatted_text(
                    HTML("<b>{metadata}</b>").format(metadata=metadata)
                )
            print(repr(pkt))
            if hasattr(pkt, "decrypted"):
                print_formatted_text(
                    HTML("<ansicyan>[i] Decrypted payload:</ansicyan>")
                )
                print(repr(pkt.decrypted))

        # Add an empty line if metadata is shown
        if show_metadata:
            print()

    # If it is not a packet, use repr method
    else:
        print(repr(pkt))

def display_event(event):
    """Display an event generated from a sniffer.
    """
    print_formatted_text(
        HTML("<ansicyan>[i] event: <b>{name}</b></ansicyan> {message}").format(
            name=event.name,
            message="("+event.message +")" if event.message is not None else ""
        )
    )

def format_analyzer_output(output, mode : str = "human_readable") -> Optional[str]:
    """Format given output depending on the provided mode.
    """
    if mode == "human_readable":
        if isinstance(output, bytes):
            return output.hex()
        if isinstance(output, str):
            return output
        return str(output)

    if mode == "raw":
        return output

    if mode == "json":
        if hasattr(output, "export_json") and callable(output.export_json):
            return output.export_json()

        if isinstance(output, bytes):
            return json.dumps(output.hex())

        try:
            return json.dumps(output)
        except TypeError:
            return None

    # Error
    return None

def wait(message, suffix : str = "", end : bool = False):
    """Display and update a waiting message.
    """
    spinner = [
        "∙∙∙",
        "●∙∙",
        "∙●∙",
        "∙∙●",
        "∙∙∙",
        "∙∙●",
        "∙●∙",
        "●∙∙",
    ]
    if hasattr(wait, "count"):
        wait.count = (wait.count + 1) % len(spinner)
    else:
        wait.count = 0
    output = f"\r\x1b[1;36m[{spinner[wait.count]}]"
    output += message + "\x1b[0;0m" + suffix
    if end:
        output = "\r" + len(output) * " " + "\x1b[#1\r"
    sys.stdout.write(output)
    sys.stdout.flush()



"""Multi-domain PCAP reader
"""
from time import sleep
from scapy.utils import rdpcap

from whad.scapy.layers import *

def patch_pcap_metadata(filename, domain):
    with open(filename, "rb") as f:
        dump = f.read()

    domain_bytes = domain.encode('ascii') + b"\x00" * (8-len(domain)) if len(domain) <= 8 else domain.encode('ascii')[:8]

    dump = dump[:8] + domain_bytes + dump[16:]
    with open(filename, "wb") as f:
        f.write(dump)

def extract_pcap_metadata(filename: str) -> str:
    """Extract metadata from a PCAP file (incompatible with PCAPng).

    :param filename: PCAP file path
    :type filename: str
    :return: Metadata
    :rtype: str
    """
    with open(filename, "rb") as f:
        dump = f.read()
    return dump[8:16].replace(b"\x00", b"").decode('ascii')

class PCAPReader(object):
    """PCAP reader class.
    """

    def __init__(self, pcapfile : str):
        """Initialize a multi-domain PCAP layer
        """
        # Load packets from PCAP file
        self.__packets = rdpcap(pcapfile)

    def packets(self, start=0, count=None, accurate=True, offset=0.0, exclude=[], filter=lambda x: True):
        """Filter packets from PCAP file and yields them.

        :param start: Start position in the PCAP
        :type start: int
        :param count: Number of packets to return
        :type count: int
        :param filter: Lambda function used to filter packets. By default, keeps everything.
        :type filter: lambda
        """
        timestamp = None
        if count is None:
            for pos, packet in enumerate(self.__packets[start:]):
                # Packet must be excluded ?
                if (pos+1) in exclude:
                    # continue
                    #timestamp = float(packet.time)
                    continue

                # Process packet
                if timestamp is None:
                    timestamp = float(packet.time)
                else:
                    delay = float(packet.time) - timestamp
                    timestamp = float(packet.time)
                    if accurate:
                        sleep(delay + offset)

                if filter(packet):
                    yield packet
        else:
            for packet in self.__packets[start:start+count]:

                # Packet must be excluded ?
                if (pos+1) in exclude:
                    # continue
                    #timestamp = float(packet.time)
                    continue

                # Process packet
                if timestamp is None:
                    timestamp = float(packet.time)
                else:
                    delay = float(packet.time) - timestamp
                    timestamp = float(packet.time)
                    if accurate:
                        sleep(delay+offset)
                if filter(packet):
                    yield packet



"""This module provides the `HIDConverter` class, an helper class to generate human-readable
keypress information from HID code/modifiers and a given locale, and to convert a specific
keycode into the corresponding HID code/modifiers for a given locale.
"""
from re import I
from whad.common.converters.hid.exceptions import (
    HIDCodeNotFound,
    HIDKeyNotFound,
    HIDLocaleNotFound,
)
from whad.common.converters.hid.mappings import HID_MAP, HID_SPECIALS


class HIDConverter:
    """
    This class provides a basic API to convert an HID code to an human friendly
    keystroke and vice versa.
    """

    @staticmethod
    def get_hid_code_from_key(
        key: str, alt=False, ctrl=False, shift=False, gui=False, locale: str = "fr"
    ) -> tuple[int, int]:
        """
        This function converts a key to a tuple composed of HID code and HID modifiers.

        :param key: main key to convert
        :param alt: alt key is pressed
        :param ctrl: ctrl key is pressed
        :param shift: shift key is pressed
        :param gui: gui key is pressed
        :param locale: Keyboard keymap to consider
        """
        if locale not in HID_MAP:
            raise HIDLocaleNotFound(locale)

        matching_codes = [m for m, k in HID_MAP[locale].items() if k == key]
        if not matching_codes:
            # Key was not found, try special keys
            matching_codes = [m for m, k in HID_SPECIALS.items() if k == key]
        if not matching_codes:
            # Key was not found in locale keymap and in special keys
            raise HIDKeyNotFound(key, alt, ctrl, shift, gui)
        # Consider first keycode found as keys are ordered from most common to rarest
        hid_code, modifiers = matching_codes[0]

        # If caller specified some modifiers, apply them
        modifier_keys = {"ALT": alt, "CTRL": ctrl, "SHIFT": shift, "GUI": gui}
        for keycode, pressed in modifier_keys.items():
            if pressed:
                mapping = [m for m, k in HID_SPECIALS.items() if k == keycode][0]
                modifiers += mapping[1]

        return (hid_code, modifiers)

    @staticmethod
    def get_key_from_hid_code(
        hid_code: int = 0, modifiers: int = 0, locale: str = "fr"
    ) -> str:
        """
        This function converts an HID code and HID modifiers to the corresponding keystroke.

        :param hid_code: HID code to convert
        :param modifiers: HID modifiers to convert
        :param locale: Keyboard keymap to consider
        """
        # Make sure the specified locale is valid
        if locale not in HID_MAP:
            raise HIDLocaleNotFound(locale)

        # Handle Ctrl, Alt and Gui keys
        if modifiers & 1:
            modifiers -= 1
            return "CTRL+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)
        if modifiers & 4:
            modifiers -= 4
            return "LALT+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)
        if modifiers & 8:
            modifiers -= 8
            return "GUI+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)
        if modifiers & 16:
            modifiers -= 16
            return "CTRL+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)
        if modifiers & 64:
            modifiers -= 64
            return "RALT+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)
        if modifiers & 128:
            modifiers -= 128
            return "RGUI+" + HIDConverter.get_key_from_hid_code(hid_code, modifiers, locale)

        # Map Right-SHIFT modifier to Left-SHIFT modifier
        if modifiers & 32:
            modifiers -= 32
            if not modifiers & 2:
                modifiers += 2

        if (hid_code, modifiers) in HID_MAP[locale]:
            return HID_MAP[locale][(hid_code, modifiers)]
        if (hid_code, modifiers) in HID_SPECIALS:
            return HID_SPECIALS[(hid_code, modifiers)]
        raise HIDCodeNotFound(hid_code, modifiers)



from scapy.packet import bind_layers, Packet
from scapy.fields import BitField, LEShortField, ByteField, StrFixedLenField, \
    FlagsField, ByteEnumField, BitEnumField
from scapy.layers.bluetooth import SM_Hdr, HCI_Event_LE_Meta, HCI_Command_Hdr, \
    HCI_Event_Command_Complete, _bluetooth_features, _bluetooth_error_codes, \
    HCI_Cmd_Complete_LE_Read_White_List_Size, ATT_Hdr

_bluetooth_supported_commands = [
    # Byte 0
    "inquiry",
    "inquiry_cancel",
    "periodic_inquiry_mode",
    "exit_periodic_inquiry_mode",
    "create_connection",
    "disconnect",
    "previously_used",
    "create_connection_cancel",

    # Byte 1
    "accept_connection_request",
    "reject_connection_request",
    "link_key_request_reply",
    "link_key_request_negative_reply",
    "pin_code_request_reply",
    "pin_code_request_negative_reply",
    "change_connection_packet_type",
    "authentication_requested",

    # Byte 2
    "set_connection_encryption",
    "change_connection_link_key",
    "link_key_selection",
    "remote_name_request",
    "remote_name_request_cancel",
    "read_remote_supported_features",
    "read_remote_extended_features",
    "read_remote_version_info",

    # Byte 3
    "read_clock_offset",
    "read_lmp_handle",
    "rfu",
    "rfu",
    "rfu",
    "rfu",
    "rfu",
    "rfu",

    # Byte 4
    "rfu",
    "hold_mode",
    "sniff_mode",
    "exit_sniff_mode",
    "previously_used",
    "previously_used",
    "qos_setup",
    "role_discovery",

    # Byte 5
    "switch_role",
    "read_link_policy_settings",
    "write_link_policy_settings",
    "read_default_link_policy_settings",
    "write_default_link_policy_settings",
    "flow_specification",
    "set_event_mask",
    "reset",

    # Byte 6
    "set_event_filter",
    "flush",
    "read_pin_type",
    "write_pin_type",
    "previously_used",
    "read_stored_link_key",
    "write_stored_link_key",
    "delete_stored_link_key",

    # Byte 7
    "write_local_name",
    "read_local_name",
    "read_connection_accept_timeout",
    "write_connection_accept_timeout",
    "read_page_timeout",
    "write_page_timeout",
    "read_scan_enable",
    "write_scan_enable",

    # Byte 8
    "read_page_scan_activity",
    "write_page_scan_activity",
    "read_inquiry_scan_activity",
    "write_inquiry_scan_activity",
    "read_authentication_enable",
    "write_authentication_enable",
    "previously_used",
    "previously_used",

    # Byte 9
    "read_class_of_device",
    "write_class_of_device",
    "read_voice_setting",
    "write_voice_setting",
    "read_automatic_flush_timeout",
    "write_automatic_flush_timeout",
    "read_num_broadcast_retransmissions",
    "write_num_broadcast_retransmissions",

    # Byte 10
    "read_hold_mode_activity",
    "write_hold_mode_activity",
    "read_transmit_power_level",
    "read_synchronous_flow_control_enable",
    "write_synchronous_flow_control_enable",
    "set_controller_to_host_flow_control",
    "host_buffer_size",
    "host_number_of_completed_packets",

    # Byte 11
    "read_link_supervision_timeout",
    "write_link_supervision_timeout",
    "read_number_of_supported_iac",
    "read_current_iac_lap",
    "write_current_iac_lap",
    "previously_used",
    "previously_used",
    "previously_used",

    # Byte 12
    "previously_used",
    "set_afh_host_channel_classification",
    "rfu",
    "rfu",
    "read_inquiry_scan_type",
    "write_inquiry_scan_type",
    "read_inquiry_mode",
    "write_inquiry_mode",

    # Byte 13
    "read_page_scan_type",
    "write_page_scan_type",
    "read_afh_channel_assessment_mode",
    "write_afh_channel_assessment_mode",
    "rfu",
    "rfu",
    "rfu",
    "rfu",

    # Byte 14
    "rfu",
    "rfu",
    "rfu",
    "read_local_version_information",
    "rfu",
    "read_local_supported_features",
    "read_local_extended_features",
    "read_buffer_size",

    # Byte 15
    "previously_used",
    "read_bd_addr",
    "read_failed_contact_counter",
    "reset_failed_contact_counter",
    "read_link_quality",
    "read_rssi",
    "read_afh_channel_map",
    "read_clock",

    # Byte 16
    "read_loopback_mode",
    "write_loopback_mode",
    "enable_device_under_test_mode",
    "setup_synchronous_connection_request",
    "accept_synchronous_connection_request",
    "reject_synchronous_connection_request",
    "rfu",
    "rfu",

    # Byte 17
    "read_extended_inquiry_response",
    "write_extended_inquiry_response",
    "refresh_encryption_key",
    "rfu",
    "sniff_subrating",
    "read_simple_pairing_mode",
    "write_simple_pairing_mode",
    "read_local_oob_data",

    # Byte 18
    "read_inquiry_response_transmit_power_level",
    "write_inquiry_transmit_power_level",
    "read_default_erroneous_data_reporting",
    "write_default_erroneous_data_reporting",
    "rfu",
    "rfu",
    "rfu",
    "io_capability_request_reply",

    # Byte 19
    "user_confirmation_request_reply",
    "user_confirmation_request_negative_reply",
    "user_passkey_request_reply",
    "user_passkey_request_negative_reply",
    "remote_oob_data_request_reply",
    "write_simple_pairing_debug_mode",
    "enhanced_flush",
    "remote_oob_data_request_negative_reply",

    # Byte 20
    "rfu",
    "rfu",
    "send_keypress_notification",
    "io_capability_request_negative_reply",
    "read_encryption_key_size",
    "rfu",
    "rfu",
    "rfu",

    # Byte 21
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",

    # Byte 22
    "previously_used",
    "previously_used",
    "set_event_mask_page_2",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",
    "previously_used",

    # Byte 23
    "read_flow_control_mode",
    "write_flow_control_mode",
    "read_data_block_size",
    "rfu",
    "rfu",
    "previously_used",
    "previously_used",
    "previously_used",

    # Byte 24
    "read_enhanced_transmit_power_level",
    "rfu",
    "previously_used",
    "previously_used",
    "previously_used",
    "read_le_host_support",
    "write_le_host_support",
    "rfu",

    # Byte 25
    "le_set_event_mask",
    "le_read_buffer_size_v1",
    "le_read_local_supported_features",
    "rfu",
    "le_set_random_address",
    "le_set_advertising_parameters",
    "le_read_advertising_physical_channel_tx_power",
    "le_set_advertising_data",

    # Byte 26
    "le_set_scan_response_data",
    "le_set_advertising_enable",
    "le_set_scan_parameters",
    "le_set_scan_enable",
    "le_create_connection",
    "le_create_connection_cancel",
    "le_read_filter_accept_list_size",
    "le_clear_filter_accept_list",

    # Byte 27
    "le_add_device_to_filter_accept_list",
    "le_remove_device_from_filter_accept_list",
    "le_connection_update",
    "le_set_host_channel_classification",
    "le_read_channel_map",
    "le_read_remote_features",
    "le_encrypt",
    "le_rand",

    # Byte 28
    "le_enable_encryption",
    "le_long_term_key_request_reply",
    "le_long_term_key_request_negative_reply",
    "le_read_supported_states",
    "le_receiver_test_v1",
    "le_transmitter_test_v1",
    "le_test_end",
    "rfu",

    # Byte 29
    "rfu",
    "rfu",
    "rfu",
    "enhanced_setup_synchronous_connection",
    "enhanced_accept_synchronous_connection",
    "read_local_supported_codecs",
    "set_mws_channel_parameters",
    "set_external_frame_configuration",

    # Byte 30
    "set_mws_signaling",
    "set_mws_transport_layer",
    "set_mws_scan_frequency_table",
    "get_mws_transport_layer_configuration",
    "set_mws_pattern_configuration",
    "set_triggered_clock_capture",
    "truncated_page",
    "truncated_page_cancel",

    # Byte 31
    "set_connectionless_peripheral_broadcast",
    "set_connectionless_peripheral_broadcast_receive",
    "start_synchronization_train",
    "receive_synchronization_train",
    "set_reserved_lt_addr",
    "delete_reserved_lt_addr",
    "set_connectionless_peripheral_broadcast_data",
    "read_synchronization_train_parameters",

    # Byte 32
    "write_synchronization_train_parameters",
    "remote_oob_extended_data_request_reply",
    "read_secure_connections_host_support",
    "write_secure_connections_host_support",
    "read_authenticated_payload_timeout",
    "write_authenticated_payload_timeout",
    "read_local_oob_extended_data",
    "write_secure_connections_test_mode",

    # Byte 33
    "read_extended_page_timeout",
    "write_extended_page_timeout",
    "read_extended_inquiry_length",
    "write_extended_inquiry_length",
    "le_remote_connection_parameter_request_reply",
    "le_remote_connection_parameter_request_negative_reply",
    "le_set_data_length",
    "le_read_suggested_default_data_length",

    # Byte 34
    "le_write_suggested_default_data_length",
    "le_read_local_p_256_public_key",
    "le_generate_dhkey_v1",
    "le_add_device_to_resolving_list",
    "le_remove_device_from_resolving_list",
    "le_clear_resolving_list",
    "le_read_resolving_list_size",
    "le_read_peer_resolvable_address",

    # Byte 35
    "le_read_local_resolvable_address",
    "le_set_address_resolution_enable",
    "le_set_resolvable_private_address_timeout",
    "le_read_maximum_data_length",
    "le_read_phy",
    "le_set_default_phy",
    "le_set_phy",
    "le_receiver_test_v2",

    # Byte 36
    "le_transmitter_test_v2",
    "le_set_advertising_set_random_address",
    "le_set_extended_advertising_parameters",
    "le_set_extended_advertising_data",
    "le_set_extended_scan_response_data",
    "le_set_extended_advertising_enable",
    "le_read_maximum_advertising_data_length",
    "le_read_number_of_supported_advertising_sets",

    # Byte 37
    "le_remove_advertising_set",
    "le_clear_advertising_sets",
    "le_set_periodic_advertising_parameters",
    "le_set_periodic_advertising_data",
    "le_set_periodic_advertising_enable",
    "le_set_extended_scan_parameters",
    "le_set_extended_scan_enable",
    "le_extended_create_connection",

    # Byte 38
    "le_periodic_advertising_create_sync",
    "le_periodic_advertising_create_sync_cancel",
    "le_periodic_advertising_terminate_sync",
    "le_add_device_to_periodic_advertiser_list",
    "le_remove_device_from_periodic_advertiser_list",
    "le_clear_periodic_advertiser_list",
    "le_read_periodic_advertiser_list_size",
    "le_read_transmit_power",

    # Byte 39
    "le_read_rf_path_compensation",
    "le_write_rf_path_compensation",
    "le_set_privacy_mode",
    "le_receiver_test_v3",
    "le_transmitter_test_v3",
    "le_set_connectionless_cte_transmit_parameters",
    "le_set_connectionless_cte_transmit_enable",
    "le_set_connectionless_iq_sampling_enable",

    # Byte 40
    "le_set_connection_cte_receive_parameters",
    "le_set_connection_cte_transmit_parameters",
    "le_connection_cte_request_enable",
    "le_connection_cte_response_enable",
    "le_read_antenna_information",
    "le_set_periodic_advertising_receive_enable",
    "le_periodic_advertising_sync_transfer",
    "le_periodic_advertising_set_info_transfer",

    # Byte 41
    "le_set_periodic_advertising_sync_transfer_parameters",
    "le_set_default_periodic_advertising_sync_transfer_parameters",
    "le_generate_dhkey_v2",
    "read_local_simple_pairing_options",
    "le_modify_sleep_clock_accuracy",
    "le_read_buffer_size_v2",
    "le_read_iso_tx_sync",
    "le_set_cig_parameters",

    # Byte 42
    "le_set_cig_parameters_test",
    "le_create_cis",
    "le_remove_cig",
    "le_accept_cis_request",
    "le_reject_cis_request",
    "le_create_big",
    "le_create_big_test",
    "le_terminate_big",

    # Byte 43
    "le_big_create_sync",
    "le_big_terminate_sync",
    "le_request_peer_sca",
    "le_setup_iso_data_path",
    "le_remove_iso_data_path",
    "le_iso_transmit_test",
    "le_iso_receive_test",
    "le_iso_read_test_counters",

    # Byte 44
    "le_iso_test_end",
    "le_set_host_feature",
    "le_read_iso_link_quality",
    "le_enhanced_read_transmit_power_level",
    "le_read_remote_transmit_power_level",
    "le_set_path_loss_reporting_parameters",
    "le_set_path_loss_reporting_enable",
    "le_set_transmit_power_reporting_enable",

    # Byte 45
    "le_transmitter_test_v4",
    "set_ecosystem_base_interval",
    "read_local_supported_codecs_v2",
    "read_local_supported_codec_capabilities",
    "read_local_supported_controller_delay",
    "configure_data_path",
    "le_set_data_related_address_changes",
    "set_min_encryption_key_size",

    # Byte 46
    "le_set_default_subrate_parameters",
    "le_subrate_request",
    "le_subrate_response",
    "le_set_periodic_advertising_sync_transfer_parameters_v2",
    "le_set_default_periodic_advertising_sync_transfer_parameters_v2",
    "le_read_transmit_power_level",
    "le_read_default_subrate_parameters",
    "le_read_max_subrate_parameters"

]

_bluetooth_le_features = [
    "encryption",                         # Bit 0
    "connection_parameters_request_procedure", # Bit 1
    "extended_reject_indication",           # Bit 2
    "peripheral_initiated_features_exchange", # Bit 3
    "ping",                               # Bit 4
    "data_packet_length_extension",       # Bit 5
    "ll_privacy",                            # Bit 6
    "extended_scanner_filter_policies",      # Bit 7
    "2m_phy",                              # Bit 8
    "stable_modulation_index_transmitter",   # Bit 9
    "stable_modulation_index_receiver",      # Bit 10
    "coded_phy",                          # Bit 11
    "extended_advertising",               # Bit 12
    "periodic_advertising",               # Bit 13
    "channel_selection_algorithm_2",         # Bit 14
    "power_class_1",                      # Bit 15
    "minimum_number_of_used_channels_procedure", # Bit 16
    "connection_cte_request",                # Bit 17
    "connection_cte_response",               # Bit 18
    "connectionless_cte_transmitter",        # Bit 19
    "connectionless_cte_receiver",           # Bit 20
    "antenna_switching_during_cte_transmission_aod", # Bit 21
    "antenna_switching_during_cte_reception_aoa",    # Bit 22
    "receiving_constant_tone_extensions",    # Bit 23
    "periodic_advertising_sync_transfer_sender",     # Bit 24
    "periodic_advertising_sync_transfer_recipient",  # Bit 25
    "sleep_clock_accuracy_updates",          # Bit 26
    "remote_public_key_validation",          # Bit 27
    "connected_isochronous_stream_central",  # Bit 28
    "connected_isochronous_stream_peripheral", # Bit 29
    "isochronous_broadcaster",               # Bit 30
    "synchronized_receiver",                 # Bit 31
    "connected_isochronous_stream",          # Bit 32
    "power_control_request",              # Bit 33
    "power_control_request",              # Bit 34
    "path_loss_monitoring",               # Bit 35
    "periodic_advertising_adi_support",      # Bit 36
    "connection_subrating",                  # Bit 37
    "connection_subrating_host_support",     # Bit 38
    "channel_classification",                # Bit 39
    "rfu",                                    # Bit 40
    "rfu",                                    # Bit 41
    "rfu",                                    # Bit 42
    "rfu",                                    # Bit 43
    "rfu",                                    # Bit 44
    "rfu",                                    # Bit 45
    "rfu",                                    # Bit 46
    "rfu",                                    # Bit 47
    "rfu",                                    # Bit 48
    "rfu",                                    # Bit 49
    "rfu",                                    # Bit 50
    "rfu",                                    # Bit 51
    "rfu",                                    # Bit 52
    "rfu",                                    # Bit 53
    "rfu",                                    # Bit 54
    "rfu",                                    # Bit 55
    "rfu",                                    # Bit 56
    "rfu",                                    # Bit 57
    "rfu",                                    # Bit 58
    "rfu",                                    # Bit 59
    "rfu",                                    # Bit 60
    "rfu",                                    # Bit 61
    "rfu",                                    # Bit 62
    "rfu",                                    # Bit 63
]

# Add missing ATT_Handle_Value_Confirmation class
class ATT_Handle_Value_Confirmation(Packet):
    """ATT Handle value confirmation packet, missing from Scapy BLE definitions
    """
    name = "Handle Value Confirmation"

class SM_Security_Request(Packet):
    name = "Security Request"
    fields_desc = [
       BitField("authentication", 0, 8)
    ]

bind_layers(SM_Hdr, SM_Security_Request, sm_command=0x0b)

class HCI_LE_Meta_Data_Length_Change(Packet):
    name = "Data Length Change"
    fields_desc = [LEShortField("handle", 0),
                   LEShortField("max_tx_octets", 0x001B),
                   LEShortField("max_tx_time", 0x0148),
                   LEShortField("max_rx_octets", 0x001B),
                   LEShortField("max_rx_time", 0x0148)
                   ]

class HCI_LE_Set_Data_Length(Packet):
    name = "Set Data Length"
    fields_desc = [LEShortField("handle", 0),
                   LEShortField("tx_octets", 0x001B),
                   LEShortField("tx_time", 0x0148),
                   ]

class HCI_Cmd_LE_Complete_Read_Buffer_Size(Packet):
    name = "LE Read Buffer Size response"
    fields_desc = [LEShortField("acl_pkt_len", 0),
                   ByteField("total_num_acl_pkts", 0)]

class HCI_Cmd_LE_Set_Event_Mask(Packet):
    name = "LE Set Event Mask"
    fields_desc = [StrFixedLenField("mask", b"\x1f\x00\x00\x00\x00\x00\x00\x00", 8)]

class HCI_Cmd_Read_Buffer_Size(Packet):
    name = "Read Buffer Size"

class HCI_Cmd_Complete_Read_Buffer_Size(Packet):
    name = "Read Buffer Size response"
    fields_desc = [LEShortField("acl_pkt_len", 0),
                   ByteField("total_num_acl_pkts", 0)]

class HCI_Cmd_Read_Local_Supported_Commands(Packet):
    name = "Read Local Supported Commands"

class HCI_Cmd_Complete_Supported_Commands(Packet):
    name = "Supported Commands response"
    fields_desc = [
        FlagsField("supported_commands", 0, -512, _bluetooth_supported_commands)
    ]

class HCI_Cmd_Read_Local_Supported_Features(Packet):
    name = "Read Local Supported Features"

class HCI_Cmd_Complete_Supported_Features(Packet):
    """
    7.7.11 Read Remote Supported Features Complete event
    """
    name = "HCI_Read_Remote_Supported_Features_Complete"
    fields_desc = [
        FlagsField('lmp_features', 0, -64, _bluetooth_features)
    ]

class HCI_Cmd_LE_Read_Local_Supported_Features(Packet):
    name = "Read Local Supported Features"

class HCI_Cmd_LE_Complete_Supported_Features(Packet):
    name = "HCI_Read_LE_Local_Supported_Features_Complete"
    fields_desc = [
        FlagsField('le_features', 0, -64, _bluetooth_le_features)
    ]

class HCI_Cmd_LE_Complete_Read_Filter_Accept_List_Size(Packet):
    name = "HCI_Read_Filter_Accept_List_Size_Complete"
    fields_desc = [
        ByteField("list_size", 0)
    ]

class HCI_Cmd_LE_Write_Suggested_Default_Data_Length(Packet):
    name = "HCI_LE_Write_Suggested_Default_Data_Length"
    fields_desc = [
        LEShortField("max_tx_octets", 0x1B),
        LEShortField("max_tx_time", 0x148),
    ]

class HCI_Cmd_LE_Read_Suggested_Default_Data_Length(Packet):
    name = "HCI_LE_Read_Suggested_Default_Data_Length"

class HCI_Cmd_LE_Complete_Suggested_Default_Data_Length(Packet):
    name = "Suggested Default Length"
    fields_desc = [
        LEShortField("max_tx_octets", 0x1B),
        LEShortField("max_tx_time", 0x148)
    ]

class HCI_Cmd_Write_Simple_Pairing_Mode(Packet):
    name = "Write Simple Pairing Mode"
    fields_desc = [ ByteField("enable", 0), ]

class HCI_Cmd_Write_Default_Link_Policy_Settings(Packet):
    name = "Write Default Link Policy Settings"
    fields_desc = [ LEShortField("policy", 7),] 

class HCI_Cmd_LE_Read_Advertising_Physical_Channel_Tx_Power(Packet):
    name = "HCI_LE_Read_Advertising_Physical_Channel_Tx_Power"
    fields_desc = []

class HCI_Cmd_Complete_LE_Advertising_Tx_Power_Level(Packet):
    name = "Advertising Tx Power Level"
    fields_desc = [ ByteField("tx_power_level", 0), ]

class HCI_Cmd_Write_Class_Of_Device(Packet):
    name = "Set Class of Device"
    fields_desc = [
        FlagsField('major_service_classes', 0, 11, [
            'limited_discoverable_mode',
            'le_audio',
            'reserved',
            'positioning',
            'networking',
            'rendering',
            'capturing',
            'object_transfer',
            'audio',
            'telephony',
            'information'
        ], tot_size=-3),
        BitEnumField('major_device_class', 0, 5, {
            0x00: 'miscellaneous',
            0x01: 'computer',
            0x02: 'phone',
            0x03: 'lan',
            0x04: 'audio_video',
            0x05: 'peripheral',
            0x06: 'imaging',
            0x07: 'wearable',
            0x08: 'toy',
            0x09: 'health',
            0x1f: 'uncategorized'
        }),
        BitField('minor_device_class', 0, 6),
        BitField('fixed', 0, 2, end_tot_size=-3)
    ]

def unbind_layer(cls, pkt_cls):
    # Unbind bottom/up
    item = None
    for guess in cls.payload_guess:
        if guess[1] == pkt_cls:
            item = guess
            break
    if item is not None:
        cls.payload_guess.remove(item)
    # Unbind top/down
    item = None
    for guess in pkt_cls.payload_guess:
        if guess[1] == cls:
            item = guess
            break
    if item is not None:
        pkt_cls.payload_guess.remove(item)

# HCI LE events
bind_layers(HCI_Command_Hdr, HCI_LE_Set_Data_Length, ogf=0x08, ocf=0x0022)
bind_layers(HCI_Command_Hdr, HCI_Cmd_Read_Buffer_Size, ogf=0x04, ocf=0x0005)

# HCI LE commands
bind_layers(HCI_Event_LE_Meta, HCI_LE_Meta_Data_Length_Change, event=7)
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Set_Event_Mask, ogf=0x08, ocf=0x0001) # noqa: E501

# HCI Commands
bind_layers(HCI_Command_Hdr, HCI_Cmd_Read_Local_Supported_Commands, ogf=0x04, ocf=0x0002) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_Read_Local_Supported_Features, ogf=0x04, ocf=0x0003) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_Write_Class_Of_Device, ogf=0x03, ocf=0x0024) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Read_Local_Supported_Features, ogf=0x08, ocf=0x0003) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Read_Advertising_Physical_Channel_Tx_Power, ogf=0x08, ocf=0x0007) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Read_Suggested_Default_Data_Length, ogf=0x08, ocf=0x0023) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Write_Suggested_Default_Data_Length, ogf=0x08, ocf=0x0024) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_Write_Simple_Pairing_Mode, ogf=0x03, ocf=0x0056) # noqa: E501
bind_layers(HCI_Command_Hdr, HCI_Cmd_Write_Default_Link_Policy_Settings, ogf=0x02, ocf=0x000f) # noqa: E501


# HCI Event Command Complete dispatch
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_LE_Complete_Read_Buffer_Size, opcode=0x2002)  # noqa: E501
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_LE_Complete_Supported_Features, opcode=0x2003)  # noqa: E501
unbind_layer(HCI_Event_Command_Complete, HCI_Cmd_Complete_LE_Read_White_List_Size)
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_LE_Complete_Read_Filter_Accept_List_Size, opcode=0x200F)  # noqa: E501
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_LE_Complete_Suggested_Default_Data_Length, opcode=0x2023)  # noqa: E501
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_LE_Advertising_Tx_Power_Level, opcode=0x2007) # noaq: E501


bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_Supported_Commands, opcode=0x1002)  # noqa: E501
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_Supported_Features, opcode=0x1003)  # noqa: E501
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_Read_Buffer_Size, opcode=0x1005)  # noqa: E501


# Bind ATT_Handle_Value_Confirmation with ATT_Hdr
bind_layers(ATT_Hdr, ATT_Handle_Value_Confirmation, opcode=0x1e)




"""
HCI configuration class.
"""

from fcntl import ioctl
from socket import socket, SOCK_RAW
from struct import pack, unpack

class HCIConfig:
    """
    This class allows to easily configure an HCI Interface.
    """
    HCIDEVDOWN = 0x400448ca
    HCIDEVRESET = 0x400448cb
    HCIDEVUP = 0x400448c9
    HCIGETDEVLIST = 0x800448d2
    HCIGETDEVINFO = 0x800448d3

    @classmethod
    def list(cls):
        """Enumerate HCI devices
        """
        sock = socket(31, SOCK_RAW, 1)

        # Get number of devices
        try:
            arg = pack('I', 16) +  b"\x00" * (8*16)
            output = ioctl(sock.fileno(), cls.HCIGETDEVLIST, arg)
            number_of_devices = unpack('H', output[:2])[0]
        except Exception:
            # Not able to open a raw HCI socket, no device available
            number_of_devices = 0

        device_ids = []
        for device_number in range(number_of_devices):
            device_id = unpack('H', output[4 + 8*device_number:4 + 8*device_number + 2])[0]
            device_ids.insert(0, device_id)

        return device_ids

    @classmethod
    def down(cls, index):
        """
        This class method stops an HCI interface.
        Its role is equivalent to the following command : ``hciconfig hci<index> down``

        :param index: index of the HCI interface to stop
        :type index: integer

        :Example:

        >>> HCIConfig.down(0)

        """

        #try:
        sock = socket(31, SOCK_RAW, 1)
        ioctl(sock.fileno(), cls.HCIDEVDOWN, index)
        sock.close()
        #except IOError:
        #    return False
        return True

    @classmethod
    def reset(cls, index):
        """
        This class method resets an HCI interface.
        Its role is equivalent to the following command : ``hciconfig hci<index> reset``

        :param index: index of the HCI interface to reset
        :type index: integer

        :Example:

        >>> HCIConfig.reset(0)

        """
        try:
            sock = socket(31, SOCK_RAW, 1)
            ioctl(sock.fileno(), cls.HCIDEVRESET, index)
            sock.close()
        except IOError:
            return False
        return True

    @classmethod
    def up(cls, index):
        """
        This class method starts an HCI interface.
        Its role is equivalent to the following command : ``hciconfig hci<index> up``

        :param index: index of the HCI interface to start
        :type index: integer

        :Example:

        >>> HCIConfig.up(0)

        """
        try:
            sock = socket(31, SOCK_RAW, 1)
            ioctl(sock.fileno(), cls.HCIDEVUP, index)
            sock.close()
        except IOError:
            return False
        return True



"""
This module provides a ::class:`whad.esb.connector.prx.PRX` connector that
implements Nordic Semiconductor's *Primary Receiver* role.

The *primary receiver* (*PRX*) role consists in receiving packets from other devices
and sending acknowledgements if required to the sending device. When a device
is in *PRX* mode, it is only capable of receiving ESB packets on a specific
channel, including *ESB pings*.

The :class:`whad.esb.connector.prx.PRX` class relies on a custom protocol stack
to handle acknowledgements and pings automatically.
"""
from typing import Generator
from scapy.packet import Packet

from whad.exceptions import UnsupportedCapability

from ..connector.base import ESB
from ..stack import ESBStack


class PRX(ESB):
    """
    Enhanced ShockBurst Primary Receiver Role (PRX) implementation for compatible WHAD device.
    
    TODO: PRX and PTX shares a lot of code in common, it could be a good idea to
    implement this code in a single class that PRX and PTX shall inherit.
    """
    def __init__(self, device):
        """Instantiate a new Primary Receiver connector.

        :param device: ESB-compatible device
        :type device: :class:`whad.device.WhadDevice`
        """
        super().__init__(device)

        # Check if device can modify its address and enter the PRX role
        self.__stack = ESBStack(self)
        self.__channel = 8
        self.__address = "11:22:33:44:55"

        self.__started = False

        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        if not self.can_be_prx():
            raise UnsupportedCapability("PrimaryReceiverMode")
        self._enable_role()

    def _enable_role(self):
        """Enable PRX role.
        """
        self.set_node_address(self.__address)
        self.enable_prx_mode(self.__channel)
        if self.__started:
            self.start()

    @property
    def stack(self) -> ESBStack:
        """Return current ESB stack instance.

        :return: Current ESB stack instance
        :rtype: :class:`whad.esb.stack.ESBStack`
        """
        return self.__stack

    @property
    def channel(self) -> int:
        """Return current channel.

        :return: current channel number
        :rtype: int
        """
        return self.__channel

    @channel.setter
    def channel(self, channel: int):
        """Set channel.

        :param channel: Channel to set
        :type channel: int
        """
        self.stop()
        self.__channel = channel
        self._enable_role()

    @property
    def address(self) -> str:
        """Return current address

        :return: ESB address
        :rtype: str
        """
        return self.__address

    @address.setter
    def address(self, address: str):
        """Set ESB address

        :param address: ESB address to set
        :type address: str
        """
        self.stop()
        self.__address = address
        self._enable_role()

    def start(self):
        """Start PRX mode.
        """
        super().start()
        self.__started = True

    def stop(self):
        """Stop PRX mode.
        """
        super().stop()
        self.__started = False

    def prepare_acknowledgment(self, ack: bytes):
        """Prepare and send acknowledgement packet

        :param ack: Acknowledgement packet to prepare
        :type ack: bytes
        """
        self.__stack.ll.prepare_acknowledgment(ack)

    def on_pdu(self, packet: Packet):
        """ESB packet reception callback

        :param packet: ESB packet received
        :type packet: :class:`scapy.packet.Packet`
        """
        self.__stack.on_pdu(packet)

    def stream(self) -> Generator[Packet, None, None]:
        """Stream received ESB packets
        """
        yield from self.__stack.ll.data_stream()



from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.zigbee.stack.apl.zdo.descriptors import SimpleDescriptor
class ApplicationObject:
    """
    This class implements an Application Object.
    In ZigBee stack, the APL layer forwards data to various applications, modelised as Application Objects.
    This class is the base class representing such an Application Object.
    """
    def __init__(
                    self,
                    name,
                    profile_id,
                    device_id,
                    device_version=0,
                    input_clusters=[],
                    output_clusters=[]
    ):
        # Reference to the manager will be populated when App is attached
        self.manager = None

        # App definition and identification
        self.name = name
        self.profile_id = profile_id
        self.device_id = device_id
        self.device_version = device_version

        # Input and output clusters, processing events
        self.input_clusters = input_clusters
        self.output_clusters = output_clusters

        # Link every cluster to the application
        for cluster in self.input_clusters + self.output_clusters:
            cluster.application = self

    @property
    def simple_descriptor(self):
        if self.manager is None:
            return None

        # Look for our own endpoint
        own_endpoint = None
        for endpoint, app in self.manager.endpoints.items():
            if app == self:
                own_endpoint = endpoint
                break

        if own_endpoint is None:
            return None

        return SimpleDescriptor(
                endpoint=own_endpoint,
                profile_identifier=self.profile_id,
                device_identifier=self.device_id,
                device_version=self.device_version,
                input_clusters=[cluster.cluster_id for cluster in app.input_clusters],
                output_clusters=[cluster.cluster_id  for cluster in app.output_clusters]
        )

    def initialize(self):
        """
        Initializes the application.
        """
        pass

    def start(self):
        """
        Starts the application.
        """
        pass

    def add_input_cluster(self, cluster):
        """
        Add an input cluster to the application.
        """
        self.input_clusters.append(cluster)
        cluster.application = self

    def add_output_cluster(self, cluster):
        """
        Add an output cluster to the application.
        """
        self.output_clusters.append(cluster)
        cluster.application = self


    def send_data(
                    self,
                    asdu,
                    destination_address_mode,
                    destination_address,
                    destination_endpoint,
                    alias_address=None,
                    alias_sequence_number=0,
                    radius=30,
                    security_enabled_transmission=False,
                    use_network_key=False,
                    acknowledged_transmission=False,
                    fragmentation_permitted=False,
                    include_extended_nonce=False,
                    cluster_id=None
    ):
        """
        Transmits a Data PDU through the APL layer.
        """
        if self.manager is None:
            return False

        return self.manager.send_data(
            asdu,
            destination_address_mode,
            destination_address,
            destination_endpoint,
            alias_address=alias_address,
            alias_sequence_number=alias_sequence_number,
            radius=radius,
            security_enabled_transmission=security_enabled_transmission,
            use_network_key=use_network_key,
            acknowledged_transmission=acknowledged_transmission,
            fragmentation_permitted=fragmentation_permitted,
            include_extended_nonce=include_extended_nonce,
            cluster_id=cluster_id,
            profile_id=self.profile_id,
            application=self
        )

    def send_interpan_data(
                            self,
                            asdu,
                            asdu_handle=0,
                            source_address_mode=MACAddressMode.SHORT,
                            destination_pan_id=0xFFFF,
                            destination_address=0xFFFF,
                            destination_address_mode=MACAddressMode.SHORT,
                            cluster_id=0,
                            acknowledged_transmission=False
    ):
        """
        Transmits an InterPAN PDU through the APL layer.
        """

        if self.manager is None:
            return False

        return self.manager.send_interpan_data(
            asdu,
            asdu_handle=asdu_handle,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            destination_address_mode=destination_address_mode,
            profile_id=self.profile_id,
            cluster_id=cluster_id,
            acknowledged_transmission=acknowledged_transmission
        )

    def on_interpan_data(
                            self,
                            asdu,
                            cluster_id=0,
                            destination_pan_id=0xFFFF,
                            destination_address=0xFFFF,
                            source_pan_id=0xFFFF,
                            source_address=0xFFFF,
                            link_quality=255
    ):
        """
        Processes an InterPAN PDU from the APL layer.

        This method forwards the InterPAN PDU to the right cluster according
        to the cluster ID.
        """
        # Iterate over all clusters
        for cluster in self.input_clusters + self.output_clusters:
            # If a cluster matches, forwards the PDU to it
            if cluster.cluster_id == cluster_id:
                cluster.on_interpan_data(
                    asdu,
                    destination_pan_id=destination_pan_id,
                    destination_address=destination_address,
                    source_pan_id=source_pan_id,
                    source_address=source_address,
                    link_quality=link_quality
                )
                return True
        return False

    def on_data(
                self,
                asdu,
                source_address,
                source_address_mode,
                cluster_id,
                security_status,
                link_quality
    ):
        """
        Processes a Data PDU from the APL layer.

        This method forwards the Data PDU to the right cluster according
        to the cluster ID.
        """
        for cluster in self.input_clusters + self.output_clusters:
            # Iterate over all clusters
            if cluster.cluster_id == cluster_id:
                # If a cluster matches, forwards the PDU to it
                cluster.on_data(
                                asdu,
                                source_address,
                                source_address_mode,
                                security_status,
                                link_quality
                )
                return True
        return False



"""
WHAD general helpers
"""

from pkgutil import iter_modules
from scapy.packet import Packet_metaclass
import whad

def message_filter(message_class):
    """Filter function to only keep messages that matches the provided class.

    :param message_class: Message class to match
    :type message_class: class
    """
    return lambda x: isinstance(x, message_class)

def is_message_type(message, category, message_type):
    """Check message type is of a given category.
    """
    if message.WhichOneof('msg') == category:
        if hasattr(message, category):
            return getattr(message, category).WhichOneof('msg') == message_type
    return False

def bd_addr_to_bytes(bd_addr):
    """
    Convert BD address to bytes
    """
    if not isinstance(bd_addr,str):
        return None

    # Clean BD address
    bd_addr_b = []
    bd_addr = bd_addr.replace(':','').lower()
    if len(bd_addr) == 12:
        for i in range(6):
            bd_addr_b.append(int(bd_addr[i*2:(i+1)*2], 16))
        return bytes(bd_addr_b[::-1])

    # Error.
    return None

def asciiz(buf: bytes) -> str:
    """Convert a bytes buffer into ascii
    """
    if not isinstance(buf, bytes):
        return None

    out=''
    for value in buf:
        if buf!=0:
            out += chr(value)
    return out

def swap_bits(value):
    """
    Swap the bits of a byte or a sequence of bytes.
    """
    # Value is an integer
    if isinstance(value, int):
        return (value * 0x0202020202 & 0x010884422010) % 1023

    # Value is of type bytes
    if isinstance(value,bytes):
        return bytes([(i * 0x0202020202 & 0x010884422010) % 1023 for i in value])

    # Error.
    return None

def bytes_to_bits(data):
    '''
    This function converts bytes to the corresponding bits sequence (as string).

    :param data: bytes to convert
    :return: corresponding bits sequence

    :Example:
    	>>> bytes_to_bits(b"\x01\x02\x03\xFF")
    	'00000001000000100000001111111111'
    	>>> bytes_to_bits(b"ABC")
    	'010000010100001001000011'
    '''
    return "".join([f"{i:08b}" for i in bytes(data)])

def bits_to_bytes(bits):
    '''
    This function converts a sequence of bits (as string) to the corresponding
    bytes.

    :param bits: string indicating a sequence of bits (e.g. "10110011")
    :return: corresponding bytes

    :Example:
    	>>> bits_to_bytes('00000001000000100000001111111111')
    	b'\x01\x02\x03\xff'
    	>>> bits_to_bytes('010000010100001001000011')
    	b'ABC'
    '''
    bitstream = [bits[i:i + 8] for i in range(0, len(bits), 8)]
    return bytes([int(j+((8-len(j))*"0"),2) for j in bitstream])

def bitwise_xor(bitseq_a, bitseq_b):
    '''
    This function returns the result of a bitwise XOR operation applied to two
    sequences of bits (a and b);

    :param a: string indicating a sequence of bits (e.g. "10101010")
    :param b: string indicating a sequence of bits (e.g. "10101010")
    :return: result of the XOR operation

    :Example:
        >>> bitwise_xor('11001111','10101010')
        '01100101'
        >>> bitwise_xor('11111111','00101010')
        '11010101'
        >>> bitwise_xor('11111111','11001100')
        '00110011'
    '''
    if len(bitseq_a) != len(bitseq_b):
        return None

    return bin(int(bitseq_a, 2)^int(bitseq_b, 2))[2:].rjust(len(bitseq_a), "0")


def list_domains():
    '''
    Returns a list of implemented domains.
    '''
    domains = []
    for submodule in iter_modules(whad.__path__):
        try:
            __import__("whad.{submodule.name}.connector")
            domains.append(submodule.name)
        except ModuleNotFoundError:
            pass
    return domains

def scapy_field_bitsize(packet, field) -> int:
    """Compute Scapy field size in bits.
    """
    if not isinstance(getattr(packet, field.name), bytes):
        return int(field.sz*8)
    return len(getattr(packet, field.name))*8


def scapy_packet_to_pattern(packet, selected_fields=None, selected_layers=None):
    '''
    This function converts a scapy packet into a pattern, a mask and an offset.
    '''

    # Format arguments to be iterable
    if isinstance(selected_fields, str):
        selected_fields = (selected_fields,)

    if isinstance(selected_layers, Packet_metaclass):
        selected_layers = (selected_layers, )

    # convert packet to bitstring
    pattern = bytes_to_bits(bytes(packet))
    mask = ""
    offset = 0

    # iterate over layers and fields to keep those selected
    for layer in packet.layers():
        use_layer = False

        # Use layer if no selected layers or layer in selected layers
        use_layer |= (selected_layers is None and selected_fields is None)
        use_layer |= (selected_layers is not None and layer in selected_layers)

        for field in layer.fields_desc:
            field_size = scapy_field_bitsize(packet, field)

            if use_layer:
                mask += "1" * field_size
            else:
                if selected_fields is None:
                    mask += "0" * field_size
                elif field.name in selected_fields:
                    mask += "1" * field_size
                else:
                    mask += "0" * field_size

    # Convert bitstrings to bytes
    pattern = bits_to_bytes(pattern)
    mask = bits_to_bytes(mask)

    # Crop leading zero bytes and adjust offset
    for mask_bit in mask:
        if mask_bit == 0:
            offset += 1
        else:
            break
    pattern = pattern[offset:]
    mask = mask[offset:]

    # Crop ending zero bytes and adjust size
    size = len(mask) - 1
    while size > 0 and mask[size] == 0:
        mask = mask[:-1]
        pattern = pattern[:-1]
        size-=1
    return (pattern, mask, offset)



"""Bluetooth Low Energy sniffing module.
"""
import logging
from typing import List, Generator
from time import sleep, time

from scapy.packet import Packet
from scapy.layers.bluetooth4LE import BTLE_DATA, BTLE

from whad.hub.ble import AccessAddressDiscovered, Synchronized, BleRawPduReceived, \
    BlePduReceived, BleAdvPduReceived

from whad.exceptions import WhadDeviceDisconnected, UnsupportedCapability, WhadDeviceNotReady
from whad.helpers import message_filter
from whad.common.sniffing import EventsManager

from .base import BLE
from .injector import Injector
from .hijacker import Hijacker
from ..exceptions import MissingCryptographicMaterial
from ..sniffing import SynchronizedConnection, SnifferConfiguration, AccessAddress, \
    SynchronizationEvent, DesynchronizationEvent, KeyExtractedEvent
from ..crypto import EncryptedSessionInitialization, LinkLayerDecryptor, \
    LegacyPairingCracking

logger = logging.getLogger(__name__)

class Sniffer(BLE, EventsManager):
    """
    BLE Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        BLE.__init__(self, device)
        EventsManager.__init__(self)
        self.__synchronized = False
        self.__connection = None
        self.__access_addresses = {}
        self.__decryptor = LinkLayerDecryptor()
        self.__encrypted_session_initialization = EncryptedSessionInitialization()
        self.__legacy_pairing_cracking = LegacyPairingCracking()
        self.__configuration = SnifferConfiguration()

        # Check if device accepts advertisements or connection sniffing
        if not self.can_sniff_advertisements() and not self.can_sniff_new_connection():
            raise UnsupportedCapability("Sniff")

    def __enter__(self) -> 'Sniffer':
        """Special method used for context management."""
        if not self.started:
            if not self.start():
                raise WhadDeviceNotReady()
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback):
        """Special method used for context management."""
        if self.started:
            if not self.stop():
                raise WhadDeviceNotReady()

    @property
    def synchronized(self):
        """Synchromization state
        """
        return self.__synchronized

    def wait_new_connection(self, address="FF:FF:FF:FF:FF:FF"):
        """Wait for synchronization on a new connection.
        """
        self.filter = address
        self.configure(advertisements=False, connection=True)
        self.start()
        while not self.synchronized:
            sleep(0.01)
        return self.__connection

    @property
    def access_address(self):
        """Access address
        """
        if self.__connection is None:
            return 0x8e89bed6
        return self.__connection.access_address

    @property
    def crc_init(self):
        """CRC seed value
        """
        if self.__connection is None:
            return 0x555555
        return self.__connection.crc_init

    @property
    def hop_interval(self):
        """Hop interval value
        """
        if self.__connection is None:
            return None
        return self.__connection.hop_interval

    @property
    def hop_increment(self):
        """Hop increment value
        """
        if self.__connection is None:
            return None
        return self.__connection.hop_interval

    @property
    def channel_map(self):
        """Channel map in use
        """
        if self.__connection is None:
            return None
        return self.__connection.channel_map

    def on_synchronized(self, access_address=None, crc_init=None, hop_increment=None,
                        hop_interval=None, channel_map=None):
        """Synchronization callback
        """
        self.__synchronized = True
        self.__connection = SynchronizedConnection(
            access_address = access_address,
            crc_init = crc_init,
            hop_increment = hop_increment,
            hop_interval = hop_interval,
            channel_map = channel_map
        )
        self.trigger_event(SynchronizationEvent(self.__connection))
        logger.info(("Connection synchronized -> access_address=0x%08x,"
                     " crc_init=0x%06x, hop_interval=%d (%d us)"
                     "hop_increment=%d, channel_map=0x%s"),
                    self.__connection.access_address,
                    self.__connection.crc_init,
                    self.__connection.hop_interval,
                    self.__connection.hop_interval*1250,
                    self.__connection.hop_increment,
                    self.__connection.channel_map.value.hex())

    def on_desynchronized(self, access_address=None):
        """Desynchronization callback
        """
        self.__synchronized = False
        self.__connection = None
        self.trigger_event(DesynchronizationEvent())
        logger.info("Connection lost.")

    def _enable_sniffing(self):
        """Enable sniffing
        """
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)

        if self.__configuration.access_addresses_discovery:
            self.discover_access_addresses()

        elif self.__configuration.active_connection is not None:

            access_address = self.__configuration.active_connection.access_address
            crc_init = self.__configuration.active_connection.crc_init
            channel_map = self.__configuration.active_connection.channel_map
            hop_interval = self.__configuration.active_connection.hop_interval
            hop_increment = self.__configuration.active_connection.hop_increment
            self.sniff_active_connection(access_address, crc_init, channel_map,
                                         hop_interval, hop_increment)

        elif self.__configuration.follow_connection:
            for key in self.__configuration.keys:
                self.__decryptor.add_key(key)

            if not self.can_sniff_new_connection():
                raise UnsupportedCapability("Sniff")

            self.sniff_new_connection(
                channel=self.__configuration.channel,
                show_advertisements=self.__configuration.show_advertisements,
                show_empty_packets=self.__configuration.show_empty_packets,
                bd_address=self.__configuration.filter
            )
        elif self.__configuration.show_advertisements:
            if not self.can_sniff_advertisements():
                raise UnsupportedCapability("Sniff")

            self.sniff_advertisements(
                channel=self.__configuration.channel,
                bd_address=self.__configuration.filter
            )


    def add_key(self, key):
        """Add known encryption key to configured keys
        """
        self.stop()
        self.__configuration.keys.append(key)
        self._enable_sniffing()

    def clear_keys(self):
        """Clear known keys
        """
        self.stop()
        self.__configuration.keys = []
        self._enable_sniffing()

    @property
    def decrypt(self) -> bool:
        """Automatic decryption status
        """
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt):
        self.stop()
        self.__configuration.decrypt = decrypt
        self._enable_sniffing()

    @property
    def configuration(self):
        """Sniffing configuration
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def configure(self, active_connection=None, access_addresses_discovery=False,
                  advertisements=True, connection=True, empty_packets=False):
        """Configure sniffer
        """
        self.stop()
        self.__configuration.active_connection = active_connection
        self.__configuration.access_addresses_discovery = access_addresses_discovery
        self.__configuration.show_advertisements = advertisements
        self.__configuration.show_empty_packets = empty_packets
        self.__configuration.follow_connection = connection
        self._enable_sniffing()

    @property
    def filter(self):
        """BD address filter
        """
        return self.__configuration.filter.upper()

    @filter.setter
    def filter(self, address="FF:FF:FF:FF:FF:FF"):
        self.stop()
        self.__configuration.filter = address.upper()
        self._enable_sniffing()

    @property
    def channel(self):
        """BLE channel number
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel=37):
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    def available_actions(self, action_filter=None) -> List[BLE]:
        """Determine the actions available once synchronized with a connection.
        """
        actions = []
        if self.__synchronized:
            if self.can_inject():
                actions.append(Injector(self.device, connection=self.__connection))

            if self.can_hijack_both() or self.can_hijack_slave() or self.can_hijack_master():
                actions.append(Hijacker(self.device, connection=self.__connection))

            return [action for action in actions
                    if action_filter is None or isinstance(action, action_filter)]

        # No action available
        return []

    def process_packet(self, packet):
        """Process sniffed packet
        """
        if self.__configuration.decrypt and BTLE_DATA in packet:
            self.__encrypted_session_initialization.process_packet(packet)
            if self.__encrypted_session_initialization.encryption:
                self.__decryptor.add_crypto_material(
                    *self.__encrypted_session_initialization.crypto_material
                )
                self.__encrypted_session_initialization.reset()
            try:
                decrypted, success = self.__decryptor.attempt_to_decrypt(packet[BTLE])
                if success:
                    #packet.decrypted = decrypted
                    decrypted_packet = decrypted
                    decrypted_packet.metadata = packet.metadata
                    packet[BTLE_DATA].remove_payload()
                    packet.payload = decrypted
                    packet.metadata = decrypted_packet.metadata
                    packet.metadata.decrypted = True
                    return packet
            except MissingCryptographicMaterial:
                pass


        if self.__configuration.pairing:
            self.__legacy_pairing_cracking.process_packet(packet[BTLE])
            if self.__legacy_pairing_cracking.ready:
                keys = self.__legacy_pairing_cracking.keys
                if keys is not None:
                    tk, stk = keys
                    logger.info("[i] New temporary key extracted: %s", tk.hex())
                    logger.info("[i] New short term key extracted: %s", stk.hex())
                    self.trigger_event(KeyExtractedEvent(stk))
                    self.__decryptor.add_key(stk)
                    self.__legacy_pairing_cracking.reset()

        return packet

    def sniff(self, timeout: float = None) -> Generator[Packet, None, None]:
        """Main sniffing function

        :param timeout: Number of seconds after which sniffing is stopped.
                        Wait forever if set to `None`.
        :type timeout: float
        """
        # Start hardware interface if not started
        if not self.started:
            if not self.start():
                raise WhadDeviceNotReady()
            stop_on_exit = True
        else:
            stop_on_exit = False

        start = time()
        try:
            if self.__configuration.access_addresses_discovery:
                for message in super().sniff(messages=(AccessAddressDiscovered), timeout=timeout):
                    if message is not None:
                        rssi = None
                        timestamp = None
                        if message.rssi:
                            rssi = message.rssi
                        if message.timestamp:
                            timestamp = message.timestamp
                        aa = message.access_address

                        if aa not in self.__access_addresses:
                            self.__access_addresses[aa] = AccessAddress(
                                aa, timestamp=timestamp,
                                rssi=rssi
                            )
                        else:
                            self.__access_addresses[aa].update(timestamp=timestamp, rssi=rssi)
                        yield self.__access_addresses[aa]

                    # Enforce timeout
                    if timeout is not None and (time() - start) > timeout:
                        break

            elif self.__configuration.active_connection is not None:
                message = self.wait_for_message(keep=message_filter(Synchronized),
                                                timeout=0.1)

                if message is not None:
                    if message.hop_increment > 0:
                        if self.support_raw_pdu():
                            message_type = BleRawPduReceived
                        elif self.__synchronized:
                            message_type = BlePduReceived
                        else:
                            message_type = BleAdvPduReceived

                        for message in super().sniff(messages=(message_type), timeout=timeout):
                            if message is not None:
                                packet = message.to_packet()
                                if packet is not None:
                                    self.monitor_packet_rx(packet)
                                    yield packet

                            # Enforce timeout
                            if timeout is not None and (time() - start) > timeout:
                                break

            else:
                if self.support_raw_pdu():
                    message_type = BleRawPduReceived
                elif self.__synchronized:
                    message_type = BlePduReceived
                else:
                    message_type = BleAdvPduReceived

                for message in super().sniff(messages=(message_type), timeout=timeout):
                    if message is not None:
                        packet = message.to_packet()
                        if packet is not None:
                            packet = self.process_packet(packet)
                            self.monitor_packet_rx(packet)
                            yield packet

                    # Enforce timeout
                    if timeout is not None and (time() - start) > timeout:
                        break


        # Handle device disconnection
        except WhadDeviceDisconnected:
            return

        # Stop current mode if we started it in the first place
        if stop_on_exit:
            if not self.stop():
                raise WhadDeviceNotReady()



from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.rf4ce.stack.apl.database import APLIB
from whad.common.stack import Layer, alias, source, state

import logging

logger = logging.getLogger(__name__)

@state(APLIB)
@alias('apl')
class APLManager(Dot15d4Manager):
    """
    This class implements the RF4CE Application manager (APL).
    It handles application-level operations and provide an high level API for network features.

    It exposes several services (e.g. "RF4CE profiles") providing a more specialized API.
    """
    def init(self):
        """
        Initialize the APL manager.

        The APL manager is associated to one or several profiles, that specialize its API
        to manipulate the lowest layers.
        """
        self.profiles = {}

    def add_profile(self, profile):
        """
        Add a profile to the APL manager.
        """
        # print("[i] Adding profile (",profile.profile_id ,profile.name ,")")
        self.profiles[profile.profile_id] = profile
        self.add_service(profile.name, profile)
        profile.attach(self)
        profile.init()

    def get_profile_by_name(self, profile_name):
        """
        Returns a profile associated to the APL manager by its name.
        """
        return self.get_service(profile_name)

    def get_profile_by_profile_id(self, profile_id):
        """
        Returns a profile associated to the APL manager by its profile ID.
        """
        if profile_id in self.profiles.keys():
            return self.profiles[profile_id]
        else:
            return None

    @source('nwk', 'NLDE-DATA')
    def on_nlde_data(self, npdu, pairing_reference, profile_id, vendor_id, link_quality, rx_flags):
        '''
        print("[i] Incoming data at APL layer", repr(npdu))
        print("    | -> pairing_reference", pairing_reference)
        print("    | -> profile_id", hex(profile_id))
        print("    | -> vendor_id", hex(vendor_id))
        print("    | -> link_quality", link_quality)
        print("    | -> rx_flags", rx_flags)
        '''
        profile = self.get_profile_by_profile_id(npdu.profile_id)
        if profile is not None:
            #print("[i] Forwarding to profile ", profile.name)
            profile.on_data(npdu, pairing_reference, vendor_id, link_quality, rx_flags)
        else:
            pass
            #print("[i] No profile found, dropping.")

    @source('nwk', 'NLME-DISCOVERY')
    def on_nlme_discovery(self, status, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, requested_device_type, link_quality):
        '''
        print("[i] Incoming discovery -> status  = ", status)
        print("    | -> source_address", hex(source_address))
        print("    | -> node_capability", node_capability)
        print("    | -> vendor_identifier", hex(vendor_identifier))
        print("    | -> vendor_string", vendor_string)
        print("    | -> application_capability", application_capability)
        print("    | -> user_string", user_string)
        print("    | -> device_type_list", device_type_list)
        print("    | -> profile_identifier_list", profile_identifier_list)
        print("    | -> requested_device_type", requested_device_type)
        print("    | -> link_quality", link_quality)
        '''
        for profile_id in profile_identifier_list:
            profile = self.get_profile_by_profile_id(profile_id)
            if profile is not None:
                #print("[i] Forwarding to profile ", profile.name)
                profile.on_discovery(status, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, requested_device_type, link_quality)
            else:
                pass
                #print("[i] No profile found, dropping.")

    @source('nwk', 'NLME-PAIR')
    def on_nlme_pair(self, status, source_pan_id, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, key_exchange_transfer_count, pairing_reference):
        '''
        print("[i] Incoming pairing -> status  = ", status)
        print("    | -> source_address", hex(source_address))
        print("    | -> node_capability", node_capability)
        print("    | -> vendor_identifier", hex(vendor_identifier))
        print("    | -> vendor_string", vendor_string)
        print("    | -> application_capability", application_capability)
        print("    | -> user_string", user_string)
        print("    | -> device_type_list", device_type_list)
        print("    | -> profile_identifier_list", profile_identifier_list)
        print("    | -> key_exchange_transfer_count", key_exchange_transfer_count)
        print("    | -> pairing_reference", pairing_reference)
        '''
        for profile_id in profile_identifier_list:
            profile = self.get_profile_by_profile_id(profile_id)
            if profile is not None:
                #print("[i] Forwarding to profile ", profile.name)
                profile.on_pair(status, source_pan_id, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, key_exchange_transfer_count, pairing_reference)
            else:
                pass
                #print("[i] No profile found, dropping.")


    @source('nwk', 'NLME-COMM-STATUS')
    def on_nlme_comm_status(self, status, pairing_reference, pan_id, destination_address_mode, destination_address):
        '''
        print("[i] Incoming communication status -> status = ", status)
        print("    | -> pairing_reference", pairing_reference)
        print("    | -> pan_id", hex(pan_id))
        print("    | -> destination_address_mode", destination_address_mode)
        print("    | -> destination_address", hex(destination_address))
        '''
        pass



"""
Pythonic LoRaWAN 1.0 stack
"""
import logging

from whad.scapy.layers.lorawan import PHYPayload
from whad.common.stack import LayerState, Layer, alias, source, state

from ..stack.llm import LWGwLinkLayer
from ..helpers import EUI

logger = logging.getLogger(__name__)

class LWGatewayState(LayerState):
    """State associated with LoRaWAN Gateway stack.
    """

    def __init__(self):
        """Initialize default state.
        """
        super().__init__()

        # Set APPKey to default value
        self.appkey = bytes.fromhex("00000000000000000000000000000000")
        self.appeui = EUI('01:02:03:04:05:06:07:08')


@alias('phy')
@state(LWGatewayState)
class LWGatewayStack(Layer):
    """LoRaWAN Gateway stack

    This stack implements the behavior of a LoRaWAN 1.0 gateway, allowing devices to join
    through OTAA or ABP.

    For now, MAC commands are not processed.
    """

    def __init__(self, connector, options={}):
        """
        Create an instance of LoRaWAN associated with a specific connector. This
        connector provides the transport layer.

        :param connector: Connector to use with this stack.
        :type connector: WhadDeviceConnector
        :param options: Stack layer options
        :type options: dict
        """

        super().__init__(options=options)

        # APPKey must be provided
        if 'appkey' in options:
            self.state.appkey = bytes.fromhex(options["appkey"])

        # APP EUI too
        if 'appeui' in options:
            self.state.appeui = EUI(options['appeui'])

        # Save connector (used as PHY layer)
        self.__connector = connector

        # Configure hardware to listen on a random uplink channel
        self.__connector.uplink()

    def get_appkey(self):
        """Retrieve the current APPKey

        :return: Current APPKey
        :rtype: bytes
        """
        return self.state.appkey
    
    def get_appeui(self):
        """Retrieve the current APP EUI

        :return: Current EUI
        :rtype: str
        """
        return self.state.appeui

    def is_device_allowed(self, dev_eui : str) -> bool:
        """Check if a device is allowed to connect.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :return: True if device is allowed to join, False otherwise.
        :rtype: bool
        """
        return self.__connector.is_device_allowed(dev_eui)

    def on_frame(self, frame : PHYPayload):
        """LoRaWAN frame callback.

        This callback handles received LoRaWAN frames.

        :param frame: LoRaWAN frame
        :type frame: PHYPayload
        """
        # Forward frame to our link layer
        self.send('ll', frame)

        # Switch back to uplink
        self.__connector.uplink()

    @source('ll')
    def send_frame(self, frame : PHYPayload, timestamp : float = None):
        """Send a LoRa frame to the current channel.

        :param frame: LoRaWAN frame to send
        :type frame: PHYPayload
        :param timestamp: Timestamp at which the payload has to be sent (if not `None`)
        :type timestamp: float
        """
        # Switch to RX1
        logger.debug("[phy] sending frame of %d bytes", len(bytes(frame)))
        self.__connector.rx1()
        self.__connector.send(bytes(frame), timestamp=timestamp)
        
    def on_device_joined(self, dev_eui : str, dev_addr : int, appskey : bytes, nwkskey : bytes):
        """A device has just joined.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param appskey: Device application session key
        :type appskey: bytes
        :param nwkskey: Device network encrytion session key
        :type nwkskey: bytes
        """
        # Forward notification to connector.
        self.__connector.on_device_joined(dev_eui, dev_addr, appskey, nwkskey)

    def on_data_received(self, dev_eui : str, dev_addr : int, data : bytes, upcount : int) -> bytes:
        """Received data from device.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param data: Data received
        :type data: bytes
        :param upcount: Uplink frame counter
        :type upcount: int
        :return: Data to send back to the device
        :rtype: bytes
        """
        return self.__connector.on_device_data(dev_eui, dev_addr, data, upcount)

LWGatewayStack.add(LWGwLinkLayer)



"""WHAD Generic events

These events are used to map specific WHAD (protobuf) messages to the corresponding
generic event for dispatching. Instances of these messages will be sent to connectors
to handle various events, independantly from the underlying wireless protocol.
"""
from typing import Any

class WhadEvent(object):
    """Basic WHAD event class

    This class stores a set of parameters and allow to access them as they were
    properties.
    """

    def __init__(self, **parameters):
        """Initialize the event and save the associated parameters.
        """
        self.__parameters = parameters

    def __getattr__(self, name: str) -> Any:
        """Access event parameters.
        """
        if name in self.__parameters:
            return self.__parameters[name]
        else:
            raise AttributeError
        
class ConnectionEvt(WhadEvent):
    """Connection Event.

    This event is sent whenever a connection to a remote device has succeeded.
    Parameters may vary depending the underlying wireless protocol.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class DisconnectionEvt(WhadEvent):
    """Disconnection event.

    This event is sent whenever an existing connection is terminated. Parameters
    may vary depending the underlying wireless protocol.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class HijackedEvt(WhadEvent):
    """Hijacked event.

    This event is sent when a connection has successfully been hijacked, no matter
    the protocol.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class TriggeredEvt(WhadEvent):
    """Triggered event.

    This event is sent when some prepared packets have been sent because of a
    met condition.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class JammedEvt(WhadEvent):
    """Jammed event.

    This event is sent to notify the success or failure of a jamming attack.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class SyncEvt(WhadEvent):
    """Synchronization event.

    This event is sent to noify that the hardware has successfully synchronized
    with an existing connection.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)

class DesyncEvt(WhadEvent):
    """Desynchronization event.

    This event is sent to notify that the current connection has been lost due
    to a desynchronization.
    """
    def __init__(self, **parameters):
        super().__init__(**parameters)


"""Bluetooth LE SMP constants
"""
SM_PAIRING_REQUEST              = 1
SM_PAIRING_RESPONSE             = 2
SM_CONFIRM                      = 3
SM_RANDOM                       = 4
SM_FAILED                       = 5
SM_ENCRYPTION_INFORMATION       = 6
SM_MASTER_IDENTIFICATION        = 7
SM_IDENTITY_INFORMATION         = 8
SM_IDENTITY_ADDRESS_INFORMATION = 9
SM_SIGNING_INFORMATION          = 0x0a
SM_SECURITY_REQUEST             = 0x0b
SM_PUBLIC_KEY                   = 0x0c
SM_DHKEY_CHECK                  = 0x0d


OOB_DISABLED = 0x00
OOB_ENABLED = 0x01

IOCAP_DISPLAY_ONLY = 0x00
IOCAP_DISPLAY_YESNO = 0x01
IOCAP_KEYBD_ONLY = 0x02
IOCAP_NOINPUT_NOOUTPUT = 0x03
IOCAP_KEYBD_DISPLAY = 0x04

PM_LEGACY_JUSTWORKS = 0x00
PM_LEGACY_PASSKEY = 0x01
PM_LEGACY_OOB = 0x02
PM_LESC_JUSTWORKS = 0x03
PM_LESC_NUMCOMP = 0x04
PM_LESC_PASSKEY = 0x05
PM_LESC_OOB = 0x06

AUTHENTICATED_METHODS = (PM_LEGACY_PASSKEY, PM_LEGACY_OOB, PM_LESC_PASSKEY, PM_LESC_NUMCOMP, PM_LESC_OOB)

SM_ERROR_PASSKEY_ENTRY_FAILED = 0x01
SM_ERROR_OOB_NOT_AVAIL = 0x02
SM_ERROR_AUTH_REQ = 0x03
SM_ERROR_CONFIRM_VALUE_FAILED = 0x04
SM_ERROR_UNSUPP_PAIRING = 0x05
SM_ERROR_ENC_KEY_SIZE = 0x06
SM_ERROR_UNSUPP_COMMAND = 0x07
SM_ERROR_UNSPEC_REASON = 0x08
SM_ERROR_REPEATED_ATTEMPTS = 0x09
SM_ERROR_INVALID_PARAMS = 0x0A
SM_ERROR_DHKEY_CHECK_FAILED = 0x0B
SM_ERROR_NUMCOMP_FAILED = 0x0C
SM_ERROR_BREDR_PAIRING = 0x0D
SM_ERROR_CROSS_TRANSP_NOT_ALLOWED = 0x0E
SM_ERROR_KEY_REJECTED = 0x0F


IOCAP_KEY_GENERATION_MAPPING = {}
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_ONLY, IOCAP_DISPLAY_ONLY)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_YESNO, IOCAP_DISPLAY_ONLY)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_ONLY, IOCAP_DISPLAY_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_NOINPUT_NOOUTPUT, IOCAP_DISPLAY_ONLY)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_DISPLAY, IOCAP_DISPLAY_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)

IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_ONLY, IOCAP_DISPLAY_YESNO)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_YESNO, IOCAP_DISPLAY_YESNO)] = (PM_LEGACY_JUSTWORKS, PM_LESC_NUMCOMP)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_ONLY, IOCAP_DISPLAY_YESNO)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_NOINPUT_NOOUTPUT, IOCAP_DISPLAY_YESNO)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_DISPLAY, IOCAP_DISPLAY_YESNO)] = (PM_LEGACY_PASSKEY, PM_LESC_NUMCOMP)

IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_ONLY, IOCAP_KEYBD_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_YESNO, IOCAP_KEYBD_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_ONLY, IOCAP_KEYBD_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_NOINPUT_NOOUTPUT, IOCAP_KEYBD_ONLY)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_DISPLAY, IOCAP_KEYBD_ONLY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)

IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_ONLY, IOCAP_NOINPUT_NOOUTPUT)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_YESNO, IOCAP_NOINPUT_NOOUTPUT)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_ONLY, IOCAP_NOINPUT_NOOUTPUT)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_NOINPUT_NOOUTPUT, IOCAP_NOINPUT_NOOUTPUT)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_DISPLAY, IOCAP_NOINPUT_NOOUTPUT)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)

IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_ONLY, IOCAP_KEYBD_DISPLAY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_DISPLAY_YESNO, IOCAP_KEYBD_DISPLAY)] = (PM_LEGACY_PASSKEY, PM_LESC_NUMCOMP)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_ONLY, IOCAP_KEYBD_DISPLAY)] = (PM_LEGACY_PASSKEY, PM_LESC_PASSKEY)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_NOINPUT_NOOUTPUT, IOCAP_KEYBD_DISPLAY)] = (PM_LEGACY_JUSTWORKS, PM_LESC_JUSTWORKS)
IOCAP_KEY_GENERATION_MAPPING[(IOCAP_KEYBD_DISPLAY, IOCAP_KEYBD_DISPLAY)] = (PM_LEGACY_PASSKEY, PM_LESC_NUMCOMP)



"""This module provides a WebSocketDevice class that can be used with a WhadDeviceConnector
to interact with a WebSocket connected to a Whad-enable device. This class implements
a WebSocket client that connects to a remote device through a WebSocket.

It also provides a dedicated connector to be used as a WebSocket server.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""
import threading
import logging
from time import sleep

from websockets.sync.server import serve
from websockets.exceptions import ConnectionClosedOK

from whad.common.converters.scapy import ScapyConverter

from .connector import Connector

logger = logging.getLogger(__name__)


class WebSocketConnector(Connector):
    """Web socket server connector.

    This connector will create a websocket and accept a single connection
    to this socket. It will then forward every message received from the
    underlying WHAD device to the connected client, and the messages received
    from the client to the device (passthrough mode).
    """

    def __init__(self, device, address="127.0.0.1", port=12345, json_mode=False):
        """Create a WebSocketConnector
        """

        # JSON mode
        self.__json_mode = json_mode

        # Input pipes
        self.__inpipe = bytearray()

        # Connected clients
        self.__clients = []
        self.__has_client = threading.Event()
        self.__no_connection = threading.Event()

        # Socket
        self.__socket = None

        # Shutdown require (to force exit in  ̀`serve()`)
        self.__shutdown_required = False

        # Initialize parent class (device connector)
        super().__init__(device)

        # Create our TCP socket
        self.__address = address
        self.__port = port

        logger.debug("Web socket address and port: %s (%s)", str(self.__address),
                     str(self.__port))


    def on_data_received(self, data: bytes):
        """Handle incoming data and parse it.

        :param data: Incoming data
        :type data: bytes
        """
        logger.debug("received raw data from socket: %s", data)
        self.__inpipe.extend(data)
        while len(self.__inpipe) > 2:
            # Is the magic correct ?
            if self.__inpipe[0] == 0xAC and self.__inpipe[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__inpipe) > 4:
                    msg_size = self.__inpipe[2] | (self.__inpipe[3] << 8)
                    if len(self.__inpipe) >= (msg_size+4):
                        raw_message = self.__inpipe[4:4+msg_size]

                        # Old parsing code
                        #_msg = Message()
                        #_msg.ParseFromString(bytes(raw_message))

                        # Parse our message with our Protocol Hub
                        _msg = self.hub.parse(bytes(raw_message))

                        # Send to device
                        if _msg is not None:
                            logger.debug(("WHAD message successfully parsed, "
                                        "forward to underlying device"))
                            self.device.send_message(_msg)

                            # Notify message
                            self.on_msg_sent(_msg)

                        # Chomp
                        self.__inpipe = self.__inpipe[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__inpipe) >= 2:
                    if (self.__inpipe[0] != 0xAC) or (self.__inpipe[1] != 0xBE):
                        self.__inpipe = self.__inpipe[1:]
                    else:
                        break

    def shutdown(self):
        """Shutdown TCP connection.
        """
        if self.__socket is not None:
            self.__socket.shutdown()


    def handler(self, websocket):
        """Handle a WebSocket client.
        
        :param websocket: websocket connection instance
        """
        logger.debug("New websocket connection on %s (%d)", websocket.remote_address[0],
                     websocket.remote_address[1])

        # Add client to our list of clients
        self.__clients.append(websocket)

        # Indicate that we have at least one client connected
        self.__has_client.set()

        # Process data from websocket
        for data in websocket:
            self.on_data_received(data)

    def serve(self, timeout: float = None):
        """Serve TCP socket

        Forward data received through our TCP socket to the underlying device.

        :param float timeout: number of seconds to wait for incoming data
        """
        logger.debug('Creating WebSocket server')

        server = serve(self.handler, self.__address, self.__port)
        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        server_thread = threading.Thread(target=server.serve_forever)
        # Exit the server thread when the main thread terminates
        server_thread.daemon = True
        server_thread.start()
        while not self.__shutdown_required:
            logger.debug("Waiting for WebSocket connection on %s (%s)",
                     self.__address, self.__port)

            # Wait for a connection on our websocket server
            if self.__has_client.wait(timeout=timeout):
                self.__has_client.clear()

                # Open device once a connection is established
                self.device.open()

                # Wait for all clients to disconnect
                if self.__no_connection.wait(timeout=timeout):
                    self.device.reset()


    # Message callbacks
    def on_any_msg(self, message):
        """Callback function to process incoming discovery messages.

        This method MUST be overriden by inherited classes.

        :param message: Discovery message
        """
        if not self.__json_mode:
            logger.debug('Received a message from device, forward to clients if any')

            clients_to_remove = []
            for client in self.__clients:
                # Convert message into bytes
                raw_message = message.serialize()

                # Define header
                header = [
                    0xAC, 0xBE,
                    len(raw_message) & 0xff,
                    (len(raw_message) >> 8) & 0xff
                ]
                try:
                    # Send header followed by serialized message
                    client.send(bytes(header))
                    client.send(raw_message)
                    logger.debug('Message sent to client')
                except ConnectionClosedOK:
                    logger.debug('Connection closed, message dropped.')
                    clients_to_remove.append(client)

            # Remove disconnected clients
            for client in clients_to_remove:
                self.__clients.remove(client)
            
            # If no client left, set the no_connection event.
            if len(self.__clients) == 0:
                self.__no_connection.set()

    def on_packet(self, packet):
        """Incoming packet handler.
        """
        if self.__json_mode:
            # Convert received packet to JSON and forwards it to connected clients
            for client in self.__clients:
                try:
                    client.send(ScapyConverter.get_json_from_scapy_packet(packet))
                except ConnectionClosedOK:
                    logger.debug('Connection closed, message dropped.')
                    
    def on_event(self, event):
        """Generic event handler.
        """

    def on_msg_sent(self, message):
        """Called when a message has successfully been sent.
        """

    def on_generic_msg(self, message):
        """Callback function to process incoming generic messages.

        This method MUST be overriden by inherited classes.

        :param message: Generic message
        :type message: HubMessage
        """

    def on_discovery_msg(self, message):
        """Callback method to process incoming discovery messages.

        :param message: Discovery message
        :type message: HubMessage
        """

    def on_domain_msg(self, domain, message):
        """Callback function to process incoming domain-related messages.

        This method MUST be overriden by inherited classes.

        :param message: Domain message
        """


"""Bluetooth Low Energy connect tool

This utility will configure a compatible whad device to connect to a given
BLE device, and chain this with another tool.
"""
import sys
import logging
from time import sleep

# WHAD device classes
from whad.device.unix import UnixConnector, UnixSocketServer
from whad.device import Bridge

# WHAD BLE interfaces
from whad.ble import Central
from whad.device.connector import LockedConnector
from whad.ble.exceptions import PeripheralNotFound

# WHAD BLE messages
from whad.hub.ble import BlePduReceived, BleRawPduReceived, Connected, \
    Disconnected, SendBlePdu, SendBleRawPdu
from whad.hub.ble.bdaddr import BDAddress

# WHAD CLI helper
from whad.cli.app import CommandLineDevicePipe, run_app

logger = logging.getLogger(__name__)

class BleConnectInputPipe(Bridge):
    """wble-connect input pipe

    When wble-connect has its standard input piped to another WHAD tool
    standard output, then it will consider that the previous tool configure
    another WHAD hardware adapter to get a connection and will process incoming
    connection/disconnection events as well as PDUs. These PDUs will be
    forwarded to the device wble-connect is connected to, and every PDU coming
    from our target device will be forwarded to the chained WHAD tool upstream.

    Inbound messages (sent by the device we are connected to) are monitored and
    only PDU-related messages will be forwarded  to the chained tool while
    outbound messages (coming from the chained tool) are analyzed to detect when
    a connection occurs and forward every PDU to our target device.
    """

    def __init__(self, input_connector, output_connector, out_handle, target_bdaddr):
        """Initialize our wble-connect input pipe.
        """
        logger.debug('[wble-connect][output-pipe] Initialization')

        # Do we get raw PDUs ?
        if output_connector.support_raw_pdu():
            self.support_raw_pdu = True
        else:
            self.support_raw_pdu = False

        # Save output handle
        self.__out_handle = out_handle
        self.__target_bdaddr = target_bdaddr

        # Initialize bridge
        super().__init__(input_connector, output_connector)

        # Unlock connectors
        output_connector.unlock(dispatch_callback=self.dispatch_pending_output_pdu)
        input_connector.unlock(dispatch_callback=self.dispatch_pending_input_pdu)

    def dispatch_pending_input_pdu(self, message):
        """Dispatch pending PDU.
        """
        logger.info("Dispatching input pdu %s", message)

        # Send message to our chained tool
        command = self.convert_packet_message(message, self.__out_handle)
        self.output.send_command(command)

    def dispatch_pending_output_pdu(self, message):
        """Dispatch pending out PDUs (received)
        """
        logger.info("Dispatching output pdu %s", message)
        self.input.send_message(message)

    def convert_packet_message(self, message, conn_handle: int):
        """Convert a BleRawPduReceived/BlePduReceived notification into the
        corresponding SendBleRawPdu/SendBlePdu command, using the provided
        connection handle.
        """
        # Do we received a packet notification ?
        logger.debug("[wble-connect][output-pipe] convert message %s into a command", message)
        if isinstance(message, BleRawPduReceived):
            # Does our input connector support raw packets ?
            if self.output.support_raw_pdu():
                logger.debug('[wble-connect][output-pipe] connector supports raw pdu')
                # Create a SendBleRawPdu command
                command = self.output.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    message.crc,
                    encrypt=False,
                    access_address=message.access_address,
                    conn_handle=conn_handle, # overwrite the connection handle
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
            else:
                logger.debug('[wble-connect][output-pipe] connector does not support raw pdu')
                # Create a SendBlePdu command
                command = self.output.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    conn_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
        elif isinstance(message, BlePduReceived):
            # Does our input connector support raw packets ?
            if self.output.support_raw_pdu():
                logger.debug('[wble-connect][output-pipe] connector supports raw pdu')
                # Create a SendBleRawPdu command
                command = self.output.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    None,
                    encrypt=False,
                    access_address=0x11223344, # We use the default access address
                    conn_handle=conn_handle, # overwrite the connection handle
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
            else:
                logger.debug('[wble-connect][output-pipe] connector does not support raw pdu')
                # Create a SendBlePdu command
                command = self.output.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    conn_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
        else:
            # Not a BLE packet notification
            command = None

        # Return generated command
        return command

    def on_inbound(self, message):
        """Process inbound messages.

        Inbound messages are coming from our output connector, i.e. the
        peripheral device we are connected to, we  only process any PDU-related
        message and discard other events/notifications such as connection or
        disconnection events.

        Normally, since we get packets from a central device we are supposed to
        be connected and know the connection handle corresponding to this
        connection.
        """
        if isinstance(message, (BleRawPduReceived, BlePduReceived)):
            logger.debug(
                "[wble-connect][output-pipe] received an inbound PDU message %s",
                message
            )
            self.input.send_message(message)
        elif isinstance(message, Disconnected):
            # Peripheral disconnected, we lock the input
            logger.debug("[wble-connect][output-pipe] Peripheral disconnected, locking input")
        elif isinstance(message, Connected):
            # We are now again connected to the target device, unlock input to
            # process any incoming message and save output connection handle
            print("Connected to %s" % self.__target_bdaddr)
            logger.debug("[wble-connect][output-pipe] Peripheral connected, updating connection handle")
            self.__out_handle = message.conn_handle
            logger.debug("[wble-connect][output-pide] Unlocking input ...")
            self.input_wrapper.unlock()
            logger.debug("[wble-connect][output-pide] Input unlocked ...")
        else:
            logger.debug("[wble-connect][output-pipe] forward default inbound message %s", message)
            # Forward other messages
            super().on_inbound(message)

    def on_outbound(self, message):
        """Process outbund messages.

        We monitor these messages to catch the connection handle used by our chained
        tool (upstream) in order to send valid commands.
        """
        if isinstance(message, (BleRawPduReceived, BlePduReceived)):
            if self.__out_handle is not None:
                logger.debug(
                    "[wble-connect][output-pipe] received an outbound PDU message %s",
                    message
                )
                command = self.convert_packet_message(message, self.__out_handle)
                self.output.send_command(command)
        elif isinstance(message, Connected):
            # A client has reconnected
            logger.error("A client has reconnected, reconnect to our target ...")
            print("Connecting to %s" % self.__target_bdaddr)
            connect = self.output.hub.ble.create_connect_to(
                self.__target_bdaddr
            )
            logger.debug("Sending connect message: %s", connect)
            self.output.send_command(connect)
        elif isinstance(message, Disconnected):
            logger.error("Client has disconnected, disconnect from device")
            self.input_wrapper.lock()

            # Create a disconnection message from scratch and send it to our output.
            disconnect = self.output.hub.ble.create_disconnect(
                self.__out_handle
            )
            self.__out_handle = None
            logger.debug("[wble-connect][output-pipe] Disconnecting target device ...")
            self.output.send_command(disconnect)
        elif isinstance(message, (SendBlePdu, SendBleRawPdu)):
            # These messages are sent by unlocked connectors that translate packets
            # into corresponding send commands
            # We just need to update the connection handle and forward them.
            message.conn_handle = self.__out_handle
            self.output.send_command(message)
        else:
            logger.debug(
                "[wble-connect][output-pipe] forward default outbound message %s",
                message
            )
            # Forward other messages
            super().on_outbound(message)

class BleConnectApp(CommandLineDevicePipe):
    """Bluetooth Low-energy connection tool

    This command-line tool has one job: initiate a BLE connection to a specific
    device and then let another tool use it for whatever purpose it has.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy connection tool',
            interface=True,
            commands=False
        )

        self.add_argument('bdaddr', metavar='BDADDR', help='Target device BD address')
        self.add_argument(
            '--spoof-public',
            metavar='PUB_BD_ADDR',
            dest='bdaddr_pub_src',
            default=None,
            help='Spoof a public BD address'
        )

        self.add_argument(
            '--spoof-random',
            metavar='RAND_BD_ADDR',
            dest='bdaddr_rand_src',
            default=None,
            help='Spoof a random BD address'
        )

        # Add an optional random type argument
        self.add_argument(
            '-r',
            '--random',
            dest='random',
            action='store_true',
            default=False,
            help='Use a random connection type'
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need a BD address to be set
            if self.args.bdaddr is not None:
                # We need to have an interface specified
                if self.interface is not None:
                    # Make sure we are piped to another tool
                    if self.is_stdout_piped() and not self.is_stdin_piped():
                        # Connect to the target device
                        self.connect_target(self.args.bdaddr, self.args.random)
                    elif self.is_stdin_piped() and not self.is_stdout_piped():
                        # Connect to the target device once the previous tool
                        # gives us a unix socket.
                        self.connect_target_and_proxify(self.args.bdaddr, self.args.random)
                    else:
                        self.error('Tool must be piped to another WHAD tool.')
                else:
                    self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt:
            self.warning("wble-connect stopped (CTL-C)")

        # Launch post-run tasks
        self.post_run()

    def set_bd_address(self, central: Central, bdaddr: str, public=True):
        """Set central BLE address
        """
        # Make sure it is a valid BD address
        if BDAddress.check(bdaddr):
            # Set the BD address
            if not central.set_bd_address(bdaddr, public=public):
                self.warning((
                    "Cannot spoof BD address, please make sure your WHAD interface "
                    "supports this feature."
                ))
        else:
            self.error(f"Invalid spoofed BD address: {bdaddr}")


    def connect_target_and_proxify(self, bdaddr, random_connection_type=False):
        """Connect to our target device and relay packets between input interface
        and our target.
        """
        # Make sure the bd address is valid
        if BDAddress.check(bdaddr):
            central = Central(self.interface)

            # Spoof source BD address if required
            if self.args.bdaddr_pub_src is not None:
                self.set_bd_address(central, self.args.bdaddr_pub_src, public=True)
            elif self.args.bdaddr_rand_src is not None:
                self.set_bd_address(central, self.args.bdaddr_rand_src, public=False)

            # Connect to our target device
            try:
                # Lock central to avoid any PDU processing
                central.lock()

                # Connect to a device
                print("Connecting to %s" % bdaddr)
                periph = central.connect(bdaddr, random_connection_type)
                print("Connected: %d" % periph.conn_handle)

                # Create our bridge: it will drive the connection process
                BleConnectInputPipe(
                    LockedConnector(self.input_interface),
                    central,
                    periph.conn_handle,
                    BDAddress(bdaddr, random=random_connection_type)
                )

                while self.input_interface.opened:
                    sleep(.2)

            except PeripheralNotFound:
                # Could not connect
                self.error(f"Cannot connect to {bdaddr}")

    def connect_target(self, bdaddr, random_connection_type=False):
        """Connect to our target device
        """
        # Make sure the bd address is valid
        if BDAddress.check(bdaddr):
            # Configure our interface (Central with PDU queue)
            logger.debug("Using CentralProxy")
            central = Central(self.interface)
            central.lock()

            # Spoof source BD address if required
            if self.args.bdaddr_pub_src is not None:
                self.set_bd_address(central, self.args.bdaddr_pub_src, public=True)
            elif self.args.bdaddr_rand_src is not None:
                self.set_bd_address(central, self.args.bdaddr_rand_src, public=False)

            # Connect to our target device
            try:
                # Output current status on stderr
                logger.info("[wble-connect::connect_target] Connecting to %s ...", bdaddr)
                sys.stderr.write(f"Connecting to {bdaddr} ...\n")
                sys.stderr.flush()

                # Connect to target device
                periph = central.connect(bdaddr, random_connection_type)

                # We are connected
                logger.info("[wble-connect::connect_target] Connected to %s", bdaddr)
                sys.stderr.write(f"Connected to {bdaddr}\n")
                sys.stderr.flush()

                # Get peers
                logger.info("local_peer: %s", central.local_peer)

                # Insanciate our output pipe
                proxy = UnixConnector(UnixSocketServer(parameters={
                    'domain':'ble',
                    'conn_handle':periph.conn_handle,
                    'initiator_bdaddr':str(central.local_peer),
                    'initiator_addrtype':str(central.local_peer.type),
                    'target_bdaddr':str(central.target_peer),
                    'target_addrtype': str(central.target_peer.type)
                }))

                # Create our custom pipe
                bridge = Bridge(central, proxy)

                # Wait for our bridge to gracefully terminate...
                bridge.join()

            except PeripheralNotFound:
                # Could not connect
                logger.info("[wble-connect::connect_target] Cannot connect to %s", bdaddr)
                self.error(f"Cannot connect to {bdaddr}")
            finally:
                central.stop()
        else:
            self.error(f"Invalid BD address: {bdaddr}")


def ble_connect_main():
    """Main application wrapper.
    """
    app = BleConnectApp()
    run_app(app)



from scapy.packet import Packet, bind_layers
from scapy.fields import ByteField, ByteEnumField, LEShortField, \
    StrLenField, FieldLenField

GOODFET_APPS = {
    0x00 : "MONITOR",
    0x51 : "CCSPI"
}

GOODFET_OPCODES = {
    0x00 : "transfer",
    0x02 : "peek",
    0x03 : "poke",
    0x80 : "read_rf_packet",
    0x81 : "send_rf_packet",
    0x84 : "peek_ram",
    0x85 : "poke_ram",
    0x10 : "setup_ccspi",
    0xB1 : "monitor_connected",
    0x7F : "init",
    0xFF : "debug_string",
    0xFE : "debug_int32",
    0xFD : "nop"
}

CC_VERSIONS = {
    0x233d : "CC2420"
}

# Scapy packets definitions
class GoodFET_Command_Hdr(Packet):
    name = "GoodFET Command header"
    fields_desc = [
        ByteEnumField("app", None, GOODFET_APPS),
        ByteEnumField("verb", None, GOODFET_OPCODES)
    ]


class GoodFET_Transfer_Command(Packet):
    name = "GoodFET Transfer Command"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]


class GoodFET_Monitor_Connected_Command(Packet):
    name = "GoodFET Monitor Connected Command"
    fields_desc = [LEShortField("size", 0x0000)]

class GoodFET_Peek_Command(Packet):
    name = "GoodFET Peek Command"
    fields_desc = [
        FieldLenField("size", None, length_of="address", fmt="<H"),
        StrLenField("address", None, length_from=lambda x:x.size)
    ]

class GoodFET_Poke_Command(Packet):
    name = "GoodFET Poke Command"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Peek_RAM_Command(Packet):
    name = "GoodFET Peek RAM Command"
    fields_desc = [
        LEShortField("size", 0x0004),
        LEShortField("address", 0x0000),
        LEShortField("count", 0x0000)
    ]

class GoodFET_Poke_RAM_Command(Packet):
    name = "GoodFET Poke RAM Command"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Send_RF_Packet_Command(Packet):
    name = "GoodFET Send RF Packet Command"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Setup_CCSPI_Command(Packet):
    name = "GoodFET Setup CCSPI Command"
    fields_desc = [LEShortField("size", 0x0000)]

class GoodFET_Read_RF_Packet_Command(Packet):
    name = "GoodFET Read RF Packet Command"
    fields_desc = [
        LEShortField("size", 0x0001),
        ByteField("reserved", 0x00) # ?
    ]

class GoodFET_Reply_Hdr(Packet):
    name = "GoodFET Reply header"
    fields_desc = [
        ByteEnumField("app", None, GOODFET_APPS),
        ByteEnumField("verb", None, GOODFET_OPCODES)
    ]


class GoodFET_Poke_RAM_Reply(Packet):
    name = "GoodFET Poke RAM Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Peek_RAM_Reply(Packet):
    name = "GoodFET Peek RAM Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Peek_Reply(Packet):
    name = "GoodFET Peek Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Init_Reply(Packet):
    name = "GoodFET Init Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="url", fmt="<H"),
        StrLenField("url", "http://goodfet.sf.net/", length_from=lambda x:x.size)
    ]

class GoodFET_Setup_CCSPI_Reply(Packet):
    name = "GoodFET Setup CCSPI Reply"
    fields_desc = [
                LEShortField("size", 0x0001),
    ]

class GoodFET_Poke_Reply(Packet):
    name = "GoodFET Poke Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Transfer_Reply(Packet):
    name = "GoodFET Transfer Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]


class GoodFET_Read_RF_Packet_Reply(Packet):
    name = "GoodFET Read RF Packet Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Send_RF_Packet_Reply(Packet):
    name = "GoodFET Send RF Packet Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

class GoodFET_Debug_String_Reply(Packet):
    name = "GoodFET Debug String Reply"
    fields_desc = [
        FieldLenField("size", None, length_of="data", fmt="<H"),
        StrLenField("data", None, length_from=lambda x:x.size)
    ]

bind_layers(GoodFET_Command_Hdr, GoodFET_Peek_Command, verb = 0x02)
bind_layers(GoodFET_Command_Hdr, GoodFET_Monitor_Connected_Command, verb = 0xB1)
bind_layers(GoodFET_Command_Hdr, GoodFET_Setup_CCSPI_Command, verb = 0x10)
bind_layers(GoodFET_Command_Hdr, GoodFET_Transfer_Command, verb = 0x00)
bind_layers(GoodFET_Command_Hdr, GoodFET_Poke_Command, verb = 0x03)
bind_layers(GoodFET_Command_Hdr, GoodFET_Read_RF_Packet_Command, verb = 0x80)
bind_layers(GoodFET_Command_Hdr, GoodFET_Send_RF_Packet_Command, verb = 0x81)
bind_layers(GoodFET_Command_Hdr, GoodFET_Peek_RAM_Command, verb = 0x84)
bind_layers(GoodFET_Command_Hdr, GoodFET_Poke_RAM_Command, verb = 0x85)

bind_layers(GoodFET_Reply_Hdr, GoodFET_Init_Reply, verb = 0x7F)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Peek_Reply, verb = 0x02)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Poke_Reply, verb = 0x03)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Debug_String_Reply, verb = 0xFF)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Setup_CCSPI_Reply, verb = 0x10)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Transfer_Reply, verb = 0x00)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Read_RF_Packet_Reply, verb = 0x80)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Send_RF_Packet_Reply, verb = 0x81)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Peek_RAM_Reply, verb = 0x84)
bind_layers(GoodFET_Reply_Hdr, GoodFET_Peek_RAM_Reply, verb = 0x85)
